import pytest

import dataclasses
import logging

import xml.etree.ElementTree as ET

from relaton_bib import BibliographicItem, DocumentRelation


@pytest.fixture
def subject():
    return DocumentRelation(
        type=DocumentRelation.Type.partOf,
        bibitem=BibliographicItem())


def test_warn_when_type_invalid(caplog):
    invalid = "invalid"

    with caplog.at_level(logging.WARNING):
        DocumentRelation(type=invalid, bibitem=None)

    assert f"invalid relation type: {invalid}" in caplog.text


def test_to_xml(subject):
    host = ET.Element("host")
    result = subject.to_xml(host)

    assert result.attrib["type"] == DocumentRelation.Type.partOf
    assert host.find("./relation/description") is None


def test_to_asciibib(subject):
    result = subject.to_asciibib()

    assert "type:: partOf" in result


def test_to_asciibib_with_pref(subject):
    result = subject.to_asciibib(prefix="tt")

    assert "tt.type:: partOf" in result


def test_hash(subject):
    result = dataclasses.asdict(subject)

    assert result["type"] == DocumentRelation.Type.partOf



from dataclasses import dataclass


@dataclass(frozen=True)
class WorkGroup:
    name: str
    number: int = None
    type: str = None
    identifier: str = None
    prefix: str = None

    # to_hash -> dataclasses.asdict

    def to_xml(self, parent):
        parent.text = self.name
        for prop in ["number", "type", "identifier", "prefix"]:
            value = getattr(self, prop)
            if value:
                parent.attrib[prop] = str(value)
        return parent

    def to_asciibib(self, prefix=""):
        pref = f"{prefix}." if prefix else prefix
        out = [f"{pref}name:: {self.name}"]
        for prop in ["number", "type", "identifier", "prefix"]:
            value = getattr(self, prop)
            if value:
                out.append(f"{pref}{prop}:: {value}")
        return "\n".join(out)



import pytest
import inspect

import dataclasses
import xml.etree.ElementTree as ET

from relaton_bib import FormattedRef, FormattedStringFormat


@pytest.fixture
def subject():
    return FormattedRef(content="content & character to escape",
                        language="en",
                        script="Latn",
                        format=FormattedStringFormat.TEXT_HTML.value)


def test_to_xml(subject):
    host = ET.Element("host")
    result = subject.to_xml(host)

    assert result.attrib["format"] == FormattedStringFormat.TEXT_HTML.value
    assert result.attrib["language"] == "en"
    assert result.attrib["script"] == "Latn"

    xmlstr = ET.tostring(result, encoding='unicode', method='xml')

    assert "content &amp;amp; character to escape" in xmlstr


def test_to_asciibib(subject):
    assert subject.to_asciibib() == inspect.cleandoc(
        """formattedref.content:: content & character to escape
           formattedref.language:: en
           formattedref.script:: Latn
           formattedref.format:: text/html""")


def test_to_hash(subject):
    result = dataclasses.asdict(subject)

    assert result["format"] == FormattedStringFormat.TEXT_HTML.value
    assert result["language"] == ["en"]
    assert result["script"] == ["Latn"]
    assert result["content"] == "content & character to escape"



import re
import logging
import xml.etree.ElementTree as ET

from dataclasses import dataclass
from enum import Enum
from typing import List

from .localized_string import LocalizedString


class BibItemLocalityType(str, Enum):
    SECTION = "section"
    CLAUSE = "clause"
    PART = "part"
    PARAGRAPH = "paragraph"
    CHAPTER = "chapter"
    PAGE = "page"
    WHOLE = "whole"
    TABLE = "table"
    ANNEX = "annex"
    FIGURE = "figure"
    NOTE = "note"
    LIST = "list"
    EXAMPLE = "example"
    VOLUME = "volume"
    ISSUE = "issue"
    TIME = "time"

    @classmethod
    def has_value(cls, value):
        return value in cls._value2member_map_


@dataclass(frozen=True)
class BibItemLocality:
    """Bibliographic item locality."""

    type: str
    reference_from: LocalizedString
    reference_to: LocalizedString = None

    def __post_init__(self):
        if not (BibItemLocalityType.has_value(self.type)
                or re.match(r"locality:[a-zA-Z0-9_]+", self.type)):
            logging.warning(
                f"[relaton-bib] invalid locality type: {self.type}")

        if isinstance(self.type, BibItemLocalityType):
            object.__setattr__(self, "type", self.type.value)

        if isinstance(self.reference_from, str):
            object.__setattr__(self, "reference_from",
                               LocalizedString(self.reference_from))

        if isinstance(self.reference_to, str):
            object.__setattr__(self, "reference_to",
                               LocalizedString(self.reference_to))

    def to_xml(self, parent):
        parent.attrib["type"] = self.type
        ET.SubElement(parent, "referenceFrom").text = str(self.reference_from)
        if self.reference_to:
            ET.SubElement(parent, "referenceTo").text = str(self.reference_to)
        return parent

    def to_asciibib(self, prefix="", count=1):
        pref = prefix + "." if prefix else prefix
        out = [f"{prefix}::"] if count > 1 else []
        out.append(f"{pref}type:: {self.type}")
        out.append(f"{pref}reference_from:: {self.reference_from}")
        if self.reference_to:
            out.append(f"{pref}reference_to:: {self.reference_to}")
        return "\n".join(out)


@dataclass(frozen=True)
class Locality(BibItemLocality):

    def to_xml(self, parent):
        node = ET.SubElement(parent, "locality")
        node.attrib["type"] = self.type
        ET.SubElement(node, "referenceFrom").text = str(self.reference_from)
        if self.reference_to:
            ET.SubElement(node, "referenceTo").text = str(self.reference_to)
        return node


@dataclass(frozen=True)
class LocalityStack:
    locality: List[Locality]

    def to_xml(self, parent):
        node = ET.SubElement(parent, "localityStack")
        for loc in self.locality:
            loc.to_xml(node)
        return node

    # TODO REVISIT and for other classes too
    # def to_hash
    #   { "locality_stack" => single_element_array(locality) }
    # end


@dataclass(frozen=True)
class SourceLocality(BibItemLocality):
    def to_xml(self, parent):
        node = ET.SubElement(parent, "sourceLocality")
        super().to_xml(node)
        return node


class SourceLocalityStack(LocalityStack):
    def to_xml(self, parent):
        node = ET.SubElement(parent, "sourceLocalityStack")
        for loc in self.locality:
            loc.to_xml(node)
        return node

#     # TODO REVISIT
#     def to_hash
#       { "source_locality_stack" => single_element_array(locality) }
#     end




from __future__ import annotations
from dataclasses import dataclass, field
from typing import List

import re
import xml.etree.ElementTree as ET

from .relaton_bib import delegate
from .document_identifier import DocumentIdType


@dataclass
@delegate("collection", "__getitem__", "__len__", "__iter__",
          "__reversed__", "__contains__")
class StructuredIdentifierCollection:
    collection: List[StructuredIdentifier]

    def to_xml(self, parent):
        for si in self.collection:
            si.to_xml(parent)

    def to_asciibib(self, prefix=""):
        pref = f"{prefix}." if prefix else prefix
        pref += "structured_identifier"
        out = []
        for si in self.collection:
            if len(self.collection) > 1:
                out.append(f"{pref}::")
            out.append(si.to_asciibib(pref))
        return "\n".join(out)

    def remove_date(self):
        for si in self.collection:
            si.remove_date()

    def remove_part(self):
        for si in self.collection:
            si.remove_part()

    def all_parts(self):
        for si in self.collection:
            si.all_parts()


@dataclass
class StructuredIdentifier:
    docnumber: str

    agency: List[str] = field(default_factory=list)

    type: str = None
    class_: str = None
    partnumber: str = None
    edition: str = None
    version: str = None
    supplementtype: str = None
    supplementnumber: str = None
    language: str = None
    year: str = None

    def to_xml(self, parent):
        name = "structuredidentifier"
        result = ET.Element(name) if parent is None \
            else ET.SubElement(parent, name)

        for a in self.agency:
            ET.SubElement(result, "agency").text = a

        if self.class_:
            ET.SubElement(result, "class").text = self.class_

        ET.SubElement(result, "docnumber").text = self.docnumber

        for opt_attr in ["partnumber", "edition", "version", "supplementtype",
                         "supplementnumber", "language", "year"]:
            value = getattr(self, opt_attr)
            if value:
                ET.SubElement(result, opt_attr).text = value
        if self.type:
            result.attrib["type"] = self.type

        return result

    def to_asciibib(self, prefix="", count=1):
        pref = f"{prefix}." if prefix else prefix
        out = [f"{pref}docnumber:: {self.docnumber}"]
        for a in self.agency:
            out.append(f"{pref}agency:: {a}")

        for opt_attr in ["type", "class", "partnumber", "edition", "version",
                         "supplementtype", "supplementnumber", "language",
                         "year"]:
            attr = opt_attr if opt_attr != "class" else "class_"
            value = getattr(self, attr)
            if value:
                out.append(f"{pref}{opt_attr}:: {value}")

        return "\n".join(out)

    def remove_date(self):
        if self.type == DocumentIdType.CN_STD:
            self.docnumber = re.sub(r"-[12]\d\d\d", "", self.docnumber)
        else:
            self.docnumber = re.sub(r":[12]\d\d\d", "", self.docnumber)
        self.year = None

    # in docid manipulations, assume ISO as the default: id-part:year
    def remove_part(self):
        self.partnumber = None
        self.docnumber = re.sub(r"-\d+", "", self.docnumber)

    def all_parts(self):
        self.docnumber += " (all parts)"



"""Unit test package for relaton_bib."""

import xml.etree.ElementTree as ET


def elements_equal(e1: ET.Element, e2: ET.Element, path: list = [], counter: dict = {}) -> bool:
    if e1.tag != e2.tag:
        print(f"{''.join(path)}| <{e1.tag} .../> != <{e2.tag} .../>")
        return False
    if e1.text != e2.text:
        if (e1.text and e2.text) and (e1.text.strip() and e2.text.strip()):
            print(f"{''.join(path)}| <{e1.tag} ...>{e1.text}</{e1.tag}> != <{e2.tag} ...>{e2.text}</{e2.tag}>")
            return False
    if e1.attrib != e2.attrib:
        diff = set(e1.attrib.keys()).symmetric_difference(e2.attrib.keys())
        if diff != {"format"}:
            print(f"{''.join(path)}| <{e1.tag} {e1.attrib} .../> != <{e2.tag} {e2.attrib} .../>")
            return False
    path.append(f"/{e1.tag}")
    if len(e1) != len(e2):
        print(f"len({''.join(path)}) {len(e1)} != {len(e2)}")
        return False
    result = all(elements_equal(c1, c2, path, dict) for c1, c2 in zip(e1, e2))
    path.pop()
    return result



from __future__ import annotations
import copy
import datetime
import logging
import re
import xml.etree.ElementTree as ET
import typing
from dataclasses import dataclass, field, is_dataclass
from enum import Enum
from typing import List

import bibtexparser
from bibtexparser.bibdatabase import BibDatabase
from bibtexparser.bwriter import BibTexWriter

from .address import Address
from .formatted_string import FormattedString
from .contribution_info import ContributionInfo, \
    ContributorRoleType
from .contact import Contact
from .contributor import Contributor
from .copyright_association import CopyrightAssociation
from .bibliographic_date import BibliographicDate, BibliographicDateType
from .series import Series, SeriesType
from .document_status import DocumentStatus
from .localized_string import LocalizedString
from .typed_title_string import TypedTitleString, TypedTitleStringCollection
from .typed_uri import TypedUri
from .formatted_ref import FormattedRef
from .medium import Medium
from .organization import Organization
from .classification import Classification
from .validity import Validity
from .bib_item_locality import BibItemLocality
from .biblio_note import BiblioNoteCollection
from .biblio_version import BibliographicItemVersion
from .place import Place
from .person import Person
from .structured_identifier import StructuredIdentifierCollection
from .editorial_group import EditorialGroup
from .ics import ICS

from .relaton_bib import to_ds_instance

from .document_relation import *
from .document_relation_collection import *

# from .bibtex_parser import BibtexPaser
# from .xml_parser import XmlPaser


class BibliographicItemType(str, Enum):
    ARTICLE = "article"
    BOOK = "book"
    BOOKLET = "booklet"
    CONFERENCE = "conference"
    MANUAL = "manual"
    PROCEEDINGS = "proceedings"
    PRESENTATION = "presentation"
    THESIS = "thesis"
    TECHREPORT = "techreport"
    STANDARD = "standard"
    UNPUBLISHED = "unpublished"
    MAP = "map"
    ELECTRONICS = "electronics"
    RESOURCE = "resource"
    AUDIOVISUAL = "audiovisual"
    FILM = "film"
    VIDEO = "video"
    BROADCAST = "broadcast"
    GRAPHIC_WORK = "graphic_work"
    MUSIC = "music"
    PATENT = "patent"
    INBOOK = "inbook"
    INCOLLECTION = "incollection"
    INPROCEEDINGS = "inproceedings"
    JOURNAL = "journal"

    @classmethod
    def has_value(cls, value):
        return value in cls._value2member_map_


@dataclass
class BibliographicItem:
    FETCHED_FORMAT: typing.ClassVar[str] = "%Y-%m-%d"

    id: str = None
    type: str = None
    docnumber: str = None
    edition: str = None
    doctype: str = None  # TODO check if Enum type better then str
    subdoctype: str = None
    title: TypedTitleStringCollection = field(
        default=TypedTitleStringCollection([]))
    link: List[TypedUri] = field(default_factory=list)
    docidentifier: List[DocumentIdentifier] = field(default_factory=list)
    date: List[BibliographicDate] = field(default_factory=list)
    abstract: List[FormattedString] = field(default_factory=list)
    contributor: List[ContributionInfo] = field(default_factory=list)
    version: BibliographicItemVersion = None
    biblionote: BiblioNoteCollection = field(default=BiblioNoteCollection([]))
    language: List[str] = field(default_factory=list)
    script: List[str] = field(default_factory=list)
    formattedref: FormattedRef = None
    status: DocumentStatus = None
    copyright: List[CopyrightAssociation] = field(default_factory=list)
    relation: DocRelationCollection = field(default=DocRelationCollection([]))
    series: List[Series] = field(default_factory=list)
    medium: Medium = None
    place: List[Place] = field(default_factory=list)
    extent: List[BibItemLocality] = field(default_factory=list)
    accesslocation: List[str] = field(default_factory=list)
    license: List[str] = field(default_factory=list)
    classification: List[Classification] = field(default_factory=list)
    validity: Validity = None
    fetched: datetime.date = None
    keyword: List[LocalizedString] = field(default_factory=list)
    editorialgroup: EditorialGroup = None
    ics: List[ICS] = field(default_factory=list)
    structuredidentifier: StructuredIdentifierCollection = None

    all_parts: bool = field(default=False, init=False, repr=False)
    _id_attribute: bool = field(default=True, init=False, repr=False)

    def __post_init__(self):
        if not BibliographicItemType.has_value(self.type):
            logging.warning(
                f"[relaton-bib] invalid document type: {self.type}")
        if isinstance(self.title, str):
            self.title = TypedTitleStringCollection([self.title])
        elif isinstance(self.title, list):
            self.title = TypedTitleStringCollection(self.title)

        self.date = list(map(to_ds_instance(BibliographicDate), self.date))

        self.abstract = list(map(to_ds_instance(FormattedString),
                                 self.abstract))

        self.copyright = list(map(to_ds_instance(CopyrightAssociation),
                                  self.copyright))

        if not self.id:
            self.id = self.makeid(None, False)

        self.link = list(map(to_ds_instance(TypedUri), self.link))
        self.place = list(map(to_ds_instance(Place), self.place))
        self.keyword = list(map(to_ds_instance(LocalizedString), self.keyword))
        if isinstance(self.type, BibliographicItemType):
            self.type = self.type.value

    def abstract_for_lang(self, lang=None):
        if lang:
            return next((a for a in self.abstract if lang in a.language), None)
        else:
            return self.abstract

    def makeid(self, docid, attribute=None):
        # TODO how this handled if not set
        # ORIGINAL CODE: return nil if attribute && !@id_attribute
        if attribute and not self._id_attribute:
            return None

        if not docid:
            docid = next((i for i in self.docidentifier if i.type != "DOI"),
                         None)
        if not docid:
            return None

        idstr = re.sub(r":", "-", docid.id)
        idstr = re.sub(r"\s", "", idstr)
        return idstr.strip()

    def shortref(self, identifier, opts={}):
        pubdate = next((d for d in self.date
                        if d.type == BibliographicDateType.PUBLISHED), None)
        year = "" if opts.get("no_year", False) or not pubdate else \
            f":{pubdate.value('on', 'year')}"
        if opts.get("all_parts") or self.all_parts:
            year += ": All Parts"

        return f"{self.makeid(identifier, False)}{year}"

    def to_xml(self, parent=None, opts={}):
        return self.render_xml(parent, opts)

    def to_bibtex(self, bibtex: BibDatabase = None) -> str:
        item = {"ENTRYTYPE": self._bibtex_type(), "ID": self.id}
        self._bibtex_title(item)
        if self.edition:
            item["edition"] = self.edition
        self._bibtex_author(item)
        self._bibtex_contributor(item)
        if any(self.place):
            item["address"] = self.place[0].name
        self._bibtex_note(item)
        self._bibtex_relation(item)
        self._bibtex_extent(item)
        self._bibtex_date(item)
        self._bibtex_series(item)
        self._bibtex_classification(item)
        if any(self.keyword):
            item["keywords"] = ", ".join(list(map(lambda k: k.content,
                                                  self.keyword)))
        self._bibtex_docidentifier(item)
        if self.fetched:
            item["timestamp"] = datetime.datetime.strftime(
                self.fetched,
                BibliographicItem.FETCHED_FORMAT)
        self._bibtex_link(item)
        if not bibtex:
            bibtex = BibDatabase()
        bibtex.entries.append(item)
        writer = BibTexWriter()
        writer.indent = '  '
        writer.common_strings = True
        writer.display_order = (
            "tile",
            "edition",
            "author",
            "publisher",
            "school",
            "organization",
            "institution",
            "address",
            "note",
            "annote",
            "howpublished",
            "comment",
            "content",
            "booktitle",
            "chapter",
            "pages",
            "volume",
            "year",
            "month",
            "urldate",
            "journal",
            "number",
            "series",
            "type",
            "mendeley-tags",
            "mendeley",
            "keywords",
            "isbn",
            "lccn",
            "issn",
            "timestamp",
            "url",
            "doi",
            "file2",
            "month_numeric")

        return bibtexparser.dumps(bibtex, writer)

    def title_for_lang(self, lang=None):
        return self.title.lang(lang)

    def url(self, link_type="src"):
        link = next((s for s in self.link if s.type == link_type), None)
        return link.content if link else None

    def disable_id_attribute(self):
        self._id_attribute = False

    def to_all_parts(self):
        """remove title part components and abstract"""
        me = copy.deepcopy(self)
        me.disable_id_attribute()
        me.relation.append(DocumentRelation(
            type=DocumentRelation.Type.instance,
            bibitem=self))
        for lang in me.language:
            me.title.delete_title_part()
            tm_en = " - ".join([t.title.content for t in me.title
                                if t.type != TypedTitleString.Type.MAIN
                                and lang in t.title.language])
            t = next((t.title for t in me.title
                      if t.type == TypedTitleString.Type.MAIN
                      and lang in t.title.language), None)
            if t:
                t.content = tm_en
        me.abstract = []
        for di in me.docidentifier:
            di.remove_part()
            di.all_parts()
            di.remove_date()
        me.structuredidentifier.remove_part()
        me.structuredidentifier.all_parts()
        me.structuredidentifier.remove_date()
        me.all_parts = True
        return me

    def to_most_recent_reference(self):
        me = copy.deepcopy(self)
        self.disable_id_attribute()
        me.relation.append(DocumentRelation(
            type=DocumentRelation.Type.instance,
            bibitem=self))
        me.abstract = []
        me.date = []
        for di in me.docidentifier:
            di.remove_date()
        si = me.structuredidentifier
        if si:
            si.remove_date()
        if me.id:
            me.id = re.sub(r"-[12]\d\d\d", "", me.id)
        return me

    def revdate(self):
        """If revision_date exists then returns it
           else returns published date or None"""
        if self.rev_date:
            return self.rev_date

        if self.version and self.version.revision_date:
            self.rev_date = self.version.revision_date
        else:
            published = next(
                (d.on for d in self.date
                    if d.type == BibliographicDateType.PUBLISHED),
                None)
            published = published
        return self.rev_date

    def to_asciibib(self, prefix=""):
        pref = f"{prefix}." if prefix else prefix
        out = [] if prefix else ["[%bibitem]", "== {blank}"]

        order = ["id", "fetched", "title", "type", "docidentifier",
                 "docnumber", "edition", "language", "script", "version",
                 "biblionote", "status", "date", "abstract", "copyright",
                 "link", "medium", "place", "extent", "accesslocation",
                 "classification", "validity", "contributor", "relation",
                 "series", "doctype", "subdoctype", "formattedref", "keyword",
                 "editorialgroup", "ics", "structuredidentifier"]

        for prop in order:
            value = getattr(self, prop)

            if isinstance(value, (str)):
                out.append(f"{pref}{prop}:: {value}")
            elif isinstance(value, datetime.date):
                out.append(
                    f"{pref}{prop}:: {value.strftime(type(self).FETCHED_FORMAT)}")
            elif hasattr(value, 'to_asciibib'):
                out.append(value.to_asciibib(prefix))
            elif hasattr(value, '__iter__'):
                if len(value) == 0:
                    continue
                if is_dataclass(value[0]):
                    p = prefix
                    if prop in ["abstract", "extent", "accesslocation",
                                "keyword"]:
                        p = f"{pref}{prop}"
                    elif prop == "contributor":
                        p = "contributor.*"
                    out += [v.to_asciibib(p, len(value)) for v in value]
                else:
                    out += [f"{pref}{prop}:: {v}" for v in value]
            elif is_dataclass(value):
                out.append(value.to_asciibib(prefix))
            elif value:
                out.append(f"{pref}{prop}:: {value}")

        return "\n".join([l for l in out if l])  # drop empty lines

    def _bibtex_title(self, item: dict):
        for t in self.title:
            if t.type == TypedTitleString.Type.MAIN:
                item["tile"] = t.title.content

    def _bibtex_type(self):
        if not self.type or self.type == BibliographicItemType.STANDARD:
            return "misc"
        return self.type

    def _bibtex_author(self, item: dict):
        def filter_authors(c):
            return isinstance(c.entity, Person) and \
                ContributorRoleType.AUTHOR in list(
                    map(lambda r: r.type, c.role))
        authors = list(
            map(lambda x: x.entity,
                list(filter(filter_authors, self.contributor))))
        if not any(authors):
            return

        authors_view = []
        for a in authors:
            if a.name.surname:
                fornames = " ".join(
                    list(map(lambda fn: str(fn), a.name.forename)))
                authors_view.append(f"{str(a.name.surname)}, {fornames}")
            else:
                authors_view.append(str(a.name.completename))

        item["author"] = " and ".join(authors_view)

    def _bibtex_contributor(self, item: dict):
        for c in self.contributor:
            role_types = list(map(lambda r: r.type, c.role))

            inst_name = None
            if ContributorRoleType.PUBLISHER in role_types:
                inst_name = "publisher"
            elif ContributorRoleType.DISTRIBUTOR in role_types:
                inst_name = None
                if self.type == BibliographicItemType.TECHREPORT:
                    inst_name = "institution"
                elif self.type in [BibliographicItemType.INPROCEEDINGS,
                                   BibliographicItemType.CONFERENCE,
                                   BibliographicItemType.MANUAL,
                                   BibliographicItemType.PROCEEDINGS]:
                    inst_name = "organization"
                elif self.type in ["mastersthesis", "phdthesis"]:
                    inst_name = "school"

            if inst_name:
                item[inst_name] = c.entity.bib_name()

    def _bibtex_note(self, item: dict):
        mapping = {
            "annote": "annote",
            "howpublished": "howpublished",
            "comment": "comment",
            "tableOfContents": "content",
            None: "note"
        }
        for n in self.biblionote:
            if n.type in mapping.keys():
                item[mapping[n.type]] = n.content

    def _bibtex_relation(self, item: dict):
        rel = next(
            (r for r in self.relation
             if r.type == DocumentRelation.Type.partOf), None)
        if rel:
            title_main = next(
                (r
                 for r in rel.bibitem.title
                 if r.type == TypedTitleString.Type.MAIN), None)
            item["booktitle"] = title_main.title.content

    def _bibtex_extent(self, item: dict):
        for e in self.extent:
            if e.type in ["chapter", "volume"]:
                item[e.type] = str(e.reference_from)
            elif e.type == "page":
                value = str(e.reference_from)
                if e.reference_to:
                    value += f"-{e.reference_to}"
                item["pages"] = value

    def _bibtex_date(self, item: dict):
        for d in self.date:
            if d.type == BibliographicDateType.PUBLISHED:
                item["year"] = str(d.value("on", "year"))
                month_num = d.value("on", "month")
                month_name = datetime.date(1900, month_num, 1).strftime('%b')
                item["month"] = month_name.lower()
                item["month_numeric"] = str(month_num)
            elif d.type == BibliographicDateType.ACCESSED:
                item["urldate"] = d.value("on")

    def _bibtex_series(self, item: dict):
        for s in self.series:
            if s.type == SeriesType.JOURNAL:
                item["journal"] = str(s.title.title.content)
                if s.number:
                    item["number"] = s.number
            elif not s.type:
                item["series"] = str(s.title.title)

    def _bibtex_classification(self, item: dict):
        mapping = {"type": "type", "mendeley": "mendeley-tags"}
        for c in self.classification:
            if c.type in mapping:
                item[mapping[c.type]] = c.value

    def _bibtex_docidentifier(self, item: dict):
        for i in self.docidentifier:
            if i.type in ["isbn", "lccn", "issn"]:
                item[i.type] = i.id

    def _bibtex_link(self, item: dict):
        mapping = {"doi": "doi", "file": "file2", "src": "url"}
        for l in self.link:
            if l.type in mapping:
                item[mapping[l.type]] = l.content

    def render_xml(self, parent: ET.Element, opts: dict) -> ET.Element:
        bibdata = opts.get("bibdata")
        lang = opts.get("lang")
        name = "bibdata" if opts.get("bibdata") else "bibitem"
        root = ET.Element(name) if parent is None \
            else ET.SubElement(parent, name)
        if self.fetched:
            ET.SubElement(root, "fetched").text = self.fetched.strftime(
                BibliographicItem.FETCHED_FORMAT)
        self.title.to_xml(root, opts)
        if self.formattedref:
            self.formattedref.to_xml(root)
        for s in self.link:
            s.to_xml(root, opts)
        for di in self.docidentifier:
            di.to_xml(root, opts)
        if self.docnumber:
            ET.SubElement(root, "docnumber").text = self.docnumber
        for d in self.date:
            d.to_xml(root, opts)
        for c in self.contributor:
            node = ET.SubElement(root, "contributor")
            for r in c.role:
                r.to_xml(node, opts)
            c.to_xml(node, opts)
        if self.edition:
            ET.SubElement(root, "edition").text = self.edition
        if self.version:
            self.version.to_xml(root)
        if self.biblionote:
            self.biblionote.to_xml(root, opts)
        if hasattr(opts.get("note"), "__iter__"):
            for n in opts["note"]:
                node = ET.SubElement(root, "note")
                node.text = n.get("text")
                node.attrib["format"] = "text/plain"
                node.attrib["type"] = n.get("type")
        for l in self.language:
            ET.SubElement(root, "language").text = l
        for s in self.script:
            ET.SubElement(root, "script").text = s
        abstr = list(filter(
            lambda ab: ab.language and lang in ab.language, self.abstract))
        abstr = abstr if any(abstr) else self.abstract
        for a in abstr:
            a.to_xml(ET.SubElement(root, "abstract"))
        if self.status:
            self.status.to_xml(root)
        for c in self.copyright:
            c.to_xml(root, opts)
        for r in self.relation:
            r.to_xml(root, opts)
        for s in self.series:
            s.to_xml(root, opts)
        if self.medium:
            self.medium.to_xml(root)
        for pl in self.place:
            pl.to_xml(root)
        for e in self.extent:
            e.to_xml(ET.SubElement(root, "extent"))
        for al in self.accesslocation:
            ET.SubElement(root, "accesslocation").text = al
        for al in self.license:
            ET.SubElement(root, "license").text = al
        for cl in self.classification:
            cl.to_xml(root)
        kwrd = list(filter(
            lambda k: k.language and lang in k.language, self.keyword))
        kwrd = kwrd if any(kwrd) else self.keyword
        for kw in kwrd:
            kw.to_xml(ET.SubElement(root, "keyword"))
        if self.validity:
            self.validity.to_xml(root)
        if opts.get("lambda"):
            opts["lambda"](root, opts)
        elif bibdata and (self.doctype or self.editorialgroup
                          or (self.ics and any(self.ics))
                          or (self.structuredIdentifier
                              and self.structuredIdentifier.presence)):
            ext = ET.SubElement(root, "ext")
            if self.doctype:
                ET.SubElement(ext, "doctype").text = self.doctype
            if self.subdoctype:
                ET.SubElement(ext, "subdoctype").text = self.subdoctype
            if self.editorialgroup:
                self.editorialgroup.to_xml(ext)
            for i in self.ics:
                i.to_xml(ext)
            if self.structuredidentifier:
                self.structuredidentifier.to_xml(ext)
        if self.id and not bibdata and not opts.get("embedded"):
            root.attrib["id"] = self.id
        if self.type:
            root.attrib["type"] = self.type
        return root

    def to_bibxml(self, parent: ET.Element = None) -> ET.Element:
        """ Render BibXML (RFC)
            parent: node where output BibXML will be built
            return root node of BibXML document"""
        reference = ET.Element("reference") if parent is None \
            else ET.SubElement(parent, "reference")
        front = ET.SubElement(reference, "front")

        reference.attrib["anchor"] = self.anchor
        if any(self.title):
            first = self.title[0].title.content
            ET.SubElement(front, "title").text = first

        self.render_seriesinfo(front)
        self.render_authors(front)
        self.render_date(front)
        self.render_workgroup(front)
        self.render_keyword(front)
        self.render_abstract(front)

        target = self.bibxml_target
        if target:
            reference.attrib["target"] = str(target.content)

        return reference

    @property
    def bibxml_target(self) -> TypedUri:
        src_link = None
        doi_link = None
        for lnk in self.link:
            if lnk.type == "src":
                src_link = lnk
                break
            elif lnk.type == "doi":
                doi_link = lnk
        return src_link or doi_link

    @property
    def anchor(self):
        did = next(
            (di for di in self.docidentifier if di.type == "rfc-anchor"),
            None)
        if did is not None:
            return did.id

        type = self.docidentifier[0].type
        return f"{type}.{self.docnumber}"

    def render_keyword(self, parent: ET.Element):
        for kw in self.keyword:
            ET.SubElement(parent, "keyword").text = kw.content

    def render_workgroup(self, parent: ET.Element):
        if self.editorialgroup is not None:
            if self.editorialgroup.technical_committee is not None:
                for tc in self.editorialgroup.technical_committee:
                    ET.SubElement(parent, "workgroup").text = tc.workgroup.name

    def render_abstract(self, parent: ET.Element):
        if not any(self.abstract):
            return

        ET.SubElement(parent, "abstract").text = re.sub(
            r"(<\/?)p(>)", r"\1t\2", self.abstract[0].content)

    def render_date(self, parent: ET.Element):
        dt = next((d for d in self.date if d.type == "published"), None)
        if dt is None:
            return

        elm = ET.SubElement(parent, "date")
        for part in ["year", "month", "day"]:
            value = dt.part(part)
            if value:
                elm.attrib[part] = month_name(value) if part == "month" \
                                                     else str(value)

    def render_seriesinfo(self, parent: ET.Element):
        for di in self.docidentifier:
            if di.type in ["DOI", "Internet-Draft"]:
                si = ET.SubElement(parent, "seriesInfo")
                si.attrib["name"] = di.type
                si.attrib["value"] = di.id

        snames = [di.type for di in self.docidentifier]
        for s in self.series:
            if s.title is None or str(s.title.title) in snames:
                continue
            si = ET.SubElement(parent, "seriesInfo")
            si.attrib["name"] = str(s.title.title)
            if s.number:
                si.attrib["value"] = s.number

    def render_authors(self, parent: ET.Element):
        for c in self.contributor:
            a = ET.SubElement(parent, "author")
            is_editor = any(r for r in c.role if r.type == "editor")
            if is_editor:
                a.attrib["role"] = "editor"

            if isinstance(c.entity, Person):
                self.render_person(a, c.entity)
            else:
                self.render_organization(a, c.entity)

            self.render_address(a, c)

    def render_address(self, parent: ET.Element, contrib: Contributor):
        addr = [cn for cn in contrib.entity.contact
                if not(isinstance(cn, Address) and cn.postcode is None)]

        if any(addr):
            address = ET.SubElement(parent, "address")
            addr = next((cn for cn in addr if isinstance(cn, Address)), None)
            postal = ET.SubElement(address, "postal")
            if addr.city:
                ET.SubElement(postal, "city").text = addr.city
            if addr.postcode:
                ET.SubElement(postal, "code").text = addr.postcode
            if addr.country:
                ET.SubElement(postal, "country").text = addr.country
            if addr.state:
                ET.SubElement(postal, "region").text = addr.state
            if any(addr.street):
                ET.SubElement(postal, "street").text = addr.street[0]

            self.render_contact(address, contrib.entity.contact)

    def render_contact(self, parent: ET.Element, addr: List[Address]):
        for t in ["phone", "email", "uri"]:
            cont = next((cn for cn in addr if isinstance(cn, Contact)
                         and cn.type == t), None)
            if cont:
                ET.SubElement(parent, t).text = cont.value

    def render_person(self, parent: ET.Element, person: Person):
        if len(person.affiliation) > 0:
            self.render_organization(parent,
                                     person.affiliation[0].organization)
        if person.name.completename is not None:
            parent.attrib["fullname"] = person.name.completename.content

        if any(person.name.initial):
            parent.attrib["initials"] = " ".join(
                [i.content for i in person.name.initial])

        if person.name.surname is not None:
            parent.attrib["surname"] = person.name.surname.content

    def render_organization(self, parent: ET.Element, org: Organization):
        o = ET.SubElement(parent, "organization")
        if len(org.name) > 0:
            o.text = org.name[0].content

        if org.abbreviation:
            o.attrib["abbrev"] = org.abbreviation.content


def month_name(month_number) -> str:
    if isinstance(month_number, str):
        month_number = int(month_number)

    return datetime.date(1900, month_number, 1).strftime('%B')



import inspect

from relaton_bib import from_bibtex, BibliographicItem


def test_parse_BibTex():
    items = from_bibtex(inspect.cleandoc(
     """@article{mrx05,
        type = "standard",
        auTHor = "Mr. X and Y, Mr.",
        editor = {Mr. Z},
        address = {Some address},
        Title = {Something Great},
        publisher = "nobody",
        YEAR = 2005,
        month = may,
        annote = {An Note},
        booktitle = {Book title},
        chapter = 4,
        edition = 2,
        howpublished = {How Published Note},
        institution = {Institution},
        journal = {Journal},
        note = {Note},
        number = 7,
        series = {Series},
        type = {Type},
        organization = {Organization},
        pages = {10-20},
        school = {School},
        volume = 1,
        urldate = {2019-12-11},
        timestamp = {2019-12-05 13:52:43},
        doi = {http://standard.org/doi-123},
        comment = {Comment},
        isbn = {isbnId},
        keywords = {Keyword, Key Word},
        language = {english},
        lccn = {lccnId},
        file2 = {file://path/file},
        mendeley-tags = {Mendeley tags},
        url = {http://standars.org/123},
        issn = {issnId},
        subtitle = {Sub title},
        content = {Content}
      },
      @mastersthesis{mrx06,
        type = "standard",
        auTHor = "Mr. X",
        address = {Some address},
        Title = {Something Great},
        publisher = "nobody",
        YEAR = 2005,
      },
      @misc{mrx07,
        type = "standard",
        auTHor = "Mr. X",
        address = {Some address},
        Title = {Something Great},
        publisher = "nobody",
        YEAR = 2005,
      },
      @conference{mrx08,
        type = "standard",
        auTHor = "Mr. X",
        address = {Some address},
        Title = {Something Great},
        publisher = "nobody",
        YEAR = 2005,
      }
    """))

    assert isinstance(items, dict)
    assert isinstance(items["mrx05"], BibliographicItem)

    # FIXME use https://stackoverflow.com/a/7060342/902217 to compare xmls
    # file = "spec/examples/from_bibtex.xml"
    # xml = items["mrx05"].to_xml
    # File.write(file, xml, encoding: "utf-8") unless File.exist? file
    # expect(xml).to be_equivalent_to File.read(file, encoding: "utf-8")



import pytest
import inspect

import dataclasses
import xml.etree.ElementTree as ET

from relaton_bib import FormattedString, FormattedStringFormat


@pytest.fixture
def formatted_str():
    return FormattedString(content="content & character to escape",
                           language="en",
                           script="Latn",
                           format=FormattedStringFormat.TEXT_HTML.value)


def test_dafault_format():
    formatted_str = FormattedString(
        "content & character to escape", "en", "Latn")
    assert formatted_str.format == FormattedStringFormat.TEXT_PLAIN.value


def test_to_xml(formatted_str):
    host = ET.Element("host")
    result = formatted_str.to_xml(host)

    assert result.attrib["format"] == FormattedStringFormat.TEXT_HTML.value
    assert result.attrib["language"] == "en"
    assert result.attrib["script"] == "Latn"

    xmlstr = ET.tostring(result, encoding='unicode', method='xml')

    assert "content &amp;amp; character to escape" in xmlstr


def test_to_asciibib(formatted_str):
    assert formatted_str.to_asciibib() == inspect.cleandoc(
        """content:: content & character to escape
        language:: en
        script:: Latn
        format:: text/html""")


def test_to_hash(formatted_str):
    result = dataclasses.asdict(formatted_str)

    assert result["format"] == FormattedStringFormat.TEXT_HTML.value
    assert result["language"] == ["en"]
    assert result["script"] == ["Latn"]
    assert result["content"] == "content & character to escape"



import dataclasses
import inspect
import pytest

import xml.etree.ElementTree as ET

from relaton_bib import ICS


@pytest.fixture
def subject():
    return ICS(code="code", text="text")


def test_to_xml(subject):
    host = ET.Element("host")
    subject.to_xml(host)

    assert host.find("./ics/code").text == "code"
    assert host.find("./ics/text").text == "text"


def test_to_asciibib(subject):
    result = subject.to_asciibib()

    assert result == inspect.cleandoc("""ics.code:: code
                                         ics.text:: text""")


def test_to_asciibib_with_pref(subject):
    result = subject.to_asciibib("p")

    assert result == inspect.cleandoc("""p.ics.code:: code
                                         p.ics.text:: text""")


def test_hash(subject):
    result = dataclasses.asdict(subject)

    assert result["code"] == "code"
    assert result["text"] == "text"



import os
import logging
import xml.etree.ElementTree as ET

from . import elements_equal
from relaton_bib import LocalityStack, SourceLocalityStack, from_xml


def test_creates_item_from_xml():
    file = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                        "examples",
                        "bib_item.xml")
    reference = ET.parse(file)

    item = from_xml(reference)

    assert item is not None
    assert elements_equal(reference.getroot(), item.to_xml())


def test_creates_item_from_bibdata_xml():
    file = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                        "examples",
                        "bibdata_item.xml")

    tree = ET.parse(file)
    reference = tree.getroot()

    item = from_xml(reference)
    assert item is not None

    result = item.to_xml(opts=dict(bibdata=True))
    assert elements_equal(reference, result)


def test_parse_date_from():
    item_xml = """<bibitem id="id">
                    <title type="main">Title</title>
                    <date type="circulated"><from>2001-02-03</from></date>
                  </bibitem>"""
    item_note = ET.fromstring(item_xml)
    item = from_xml(item_note)
    assert item.date[0].value("from_") == "2001-02-03"


def test_parse_locality_not_inclosed_in_localityStack():
    item_xml = """
        <bibitem id="id">
            <title type="main">Title</title>
            <relation type="updates">
            <bibitem>
                <formattedref format="text/plain">ISO 19115</formattedref>
            </bibitem>
            <locality type="section">
                <referenceFrom>Reference from</referenceFrom>
            </locality>
            </relation>
        </bibitem>"""
    item_note = ET.fromstring(item_xml)
    item = from_xml(item_note)
    assert type(item.relation[0].locality[0]) is LocalityStack


def test_parse_sourceLocality_not_inclosed_in_sourceLocalityStack():
    item_xml = """
        <bibitem id="id">
          <title type="main">Title</title>
          <relation type="updates">
            <bibitem>
              <formattedref format="text/plain">ISO 19115</formattedref>
            </bibitem>
            <sourceLocality type="section">
              <referenceFrom>Reference from</referenceFrom>
            </sourceLocality>
          </relation>
        </bibitem>"""
    item_note = ET.fromstring(item_xml)
    item = from_xml(item_note)
    assert type(item.relation[0].source_locality[0]) is SourceLocalityStack


def test_ignore_empty_dates():
    item_xml = """
        <bibitem id="id">
            <title type="main">Title</title>
            <date type="circulated" />
        </bibitem>"""
    item_note = ET.fromstring(item_xml)
    item = from_xml(item_note)
    assert len(item.date) == 0


def test_warn_if_XML_empty(caplog):
    item = None
    with caplog.at_level(logging.WARNING):
        item = from_xml(ET.Element("test"))

    assert "can't find bibitem" in caplog.text
    assert item is None



import dataclasses
import datetime
import inspect
import pytest

import xml.etree.ElementTree as ET

from relaton_bib import Validity


@pytest.fixture
def subject():
    return Validity(
        begins=datetime.datetime(2009, 1, 1, 0, 0),
        ends=datetime.datetime(2012, 11, 28, 1, 2))


def test_to_xml(subject):
    host = ET.Element("host")
    result = subject.to_xml(host)

    assert host.find("./validity") == result
    assert result.find("./validityBegins").text == "2009-01-01 00:00"
    assert result.find("./validityEnds").text == "2012-11-28 01:02"


def test_to_asciibib(subject):
    result = subject.to_asciibib()

    assert result == inspect.cleandoc(
        """validity.begins:: 2009-01-01 00:00
           validity.ends:: 2012-11-28 01:02""")


def test_to_asciibib_with_pref(subject):
    result = subject.to_asciibib(prefix="eee")

    assert result == inspect.cleandoc(
        """eee.validity.begins:: 2009-01-01 00:00
           eee.validity.ends:: 2012-11-28 01:02""")


def test_hash(subject):
    result = dataclasses.asdict(subject)

    assert type(result["begins"]) is datetime.datetime
    assert type(result["ends"]) is datetime.datetime



import pytest

import dataclasses
import inspect
import logging
import xml.etree.ElementTree as ET

from relaton_bib import Series, TypedTitleString
from relaton_bib import FormattedRef
from relaton_bib import LocalizedString


@pytest.fixture
def subject_min():
    return Series(type="main", title=TypedTitleString(content="value"))


def test_invalid_type(caplog):
    invalid = "invalid"

    with caplog.at_level(logging.WARNING):
        Series(type=invalid, title=TypedTitleString(content="value"))

    assert f"[relaton-bib] Series type is invalid: {invalid}" in caplog.text


def test_missing_title_and_formattedref():
    with pytest.raises(ValueError) as excinfo:
        Series(type="alt")
    assert "arg `title` or `formattedref` should present" in str(excinfo.value)


def test_to_xml(subject_min):
    host = ET.Element("host")
    result = subject_min.to_xml(host)

    assert host.find("./series") is not None
    assert result.attrib["type"] == "main"
    assert result.find("./title").text == "value"
    assert result.find("./title").attrib["format"] == "text/plain"


def test_to_asciibib(subject_min):
    result = subject_min.to_asciibib()

    assert result == inspect.cleandoc(
        """series.type:: main
           series.title.content:: value
           series.title.format:: text/plain""")


def test_to_asciibib_with_pref(subject_min):
    result = subject_min.to_asciibib(prefix="test")

    assert result == inspect.cleandoc(
        """test.series.type:: main
           test.series.title.content:: value
           test.series.title.format:: text/plain""")


def test_hash(subject_min):
    result = dataclasses.asdict(subject_min)

    assert result["type"] == "main"


@pytest.fixture
def subject_full():
    return Series(
        type="main",
        formattedref=FormattedRef("abbrv", "en", "Latn"),
        place="some_place",
        organization="org",
        abbreviation=LocalizedString("abbrv", "en", "Latn"),
        from_="from",
        to="to",
        number="123",
        partnumber="321")


def test_full_to_xml(subject_full):
    host = ET.Element("host")
    result = subject_full.to_xml(host)

    assert host.find("./series") is not None
    assert result.attrib["type"] == "main"
    assert result.find("./title") is None


def test_full_to_asciibib(subject_full):
    result = subject_full.to_asciibib()

    assert result == inspect.cleandoc(
        """series.type:: main
           series.formattedref.content:: abbrv
           series.formattedref.language:: en
           series.formattedref.script:: Latn
           series.formattedref.format:: text/plain
           series.place:: some_place
           series.organization:: org
           series.abbreviation.content:: abbrv
           series.abbreviation.language:: en
           series.abbreviation.script:: Latn
           series.from:: from
           series.to:: to
           series.number:: 123
           series.partnumber:: 321""")


def test_full_to_asciibib_with_pref(subject_full):
    result = subject_full.to_asciibib(prefix="pref")

    assert result == inspect.cleandoc(
        """pref.series.type:: main
           pref.series.formattedref.content:: abbrv
           pref.series.formattedref.language:: en
           pref.series.formattedref.script:: Latn
           pref.series.formattedref.format:: text/plain
           pref.series.place:: some_place
           pref.series.organization:: org
           pref.series.abbreviation.content:: abbrv
           pref.series.abbreviation.language:: en
           pref.series.abbreviation.script:: Latn
           pref.series.from:: from
           pref.series.to:: to
           pref.series.number:: 123
           pref.series.partnumber:: 321""")


def test_full_hash(subject_full):
    result = dataclasses.asdict(subject_full)

    assert result["type"] == "main"
    assert result["place"] == "some_place"
    assert result["from_"] == "from"
    assert result["to"] == "to"
    assert result["number"] == "123"
    assert result["partnumber"] == "321"



from dataclasses import dataclass, field
from typing import List

import xml.etree.ElementTree as ET


@dataclass
class Address:
    city: str
    country: str
    street: List[str] = field(default_factory=list)
    state: str = None
    postcode: str = None

    def to_xml(self, parent):
        name = "address"
        result = ET.Element(name) if parent is None \
            else ET.SubElement(parent, name)
        for st in self.street:
            ET.SubElement(result, "street").text = st
        ET.SubElement(result, "city").text = self.city
        if self.state:
            ET.SubElement(result, "state").text = self.state
        ET.SubElement(result, "country").text = self.country
        if self.postcode:
            ET.SubElement(result, "postcode").text = self.postcode
        return result

    def to_asciibib(self, prefix="", count=1):
        pref = f"{prefix}.address" if prefix else "address"
        out = [f"{pref}::"] if count > 1 else []
        for st in self.street:
            out.append(f"{pref}.street:: {st}")
        out.append(f"{pref}.city:: {self.city}")
        if self.state:
            out.append(f"{pref}.state:: {self.state}")
        out.append(f"{pref}.country:: {self.country}")
        if self.postcode:
            out.append(f"{pref}.postcode:: {self.postcode}")
        return "\n".join(out)



from __future__ import annotations
from dataclasses import dataclass, field
from typing import List, Union

from .address import Address
from .contact import Contact


@dataclass
class Contributor:
    uri: str = None
    contact: List[Union[Address, Contact]] = field(default_factory=list)

    @property
    def url(self):
        return self.uri

    def to_xml(self, parent):
        for c in self.contact:
            c.to_xml(parent)
        return parent

    def to_asciibib(self, prefix=""):
        pref = f"{prefix}." if prefix else prefix
        out = []
        if self.uri:
            out.append(f"{pref}url:: {self.uri}")
        addr = [x for x in self.contact if isinstance(x, Address)]
        for a in addr:
            out.append(a.to_asciibib(prefix, len(addr)))
        cont = [x for x in self.contact if isinstance(x, Contact)]
        for c in cont:
            out.append(c.to_asciibib(prefix, len(cont)))
        return "\n".join(out)

    def bib_name(self) -> str:
        raise NotImplementedError()



import dataclasses
import inspect
import pytest

import xml.etree.ElementTree as ET

from relaton_bib import WorkGroup, TechnicalCommittee


@pytest.fixture
def subject():
    return TechnicalCommittee(WorkGroup(name="value",
                                        number=1,
                                        type="type"))


def test_to_xml(subject):
    host = ET.Element("host")
    result = subject.to_xml(host)

    assert host.find("./technical-committee") is not None
    assert result.text == "value"
    assert result.attrib["number"] == "1"


def test_to_asciibib(subject):
    result = subject.to_asciibib()

    assert result == inspect.cleandoc(
        """technical_committee.name:: value
           technical_committee.number:: 1
           technical_committee.type:: type""")


def test_to_asciibib_with_pref(subject):
    result = subject.to_asciibib(prefix="a")

    assert result == inspect.cleandoc(
        """a.technical_committee.name:: value
           a.technical_committee.number:: 1
           a.technical_committee.type:: type""")


def test_hash(subject):
    result = dataclasses.asdict(subject)

    assert result["workgroup"]["type"] == "type"
    assert result["workgroup"]["number"] == 1
    assert result["workgroup"]["name"] == "value"



from dataclasses import dataclass
from enum import Enum

import logging
import xml.etree.ElementTree as ET

from .formatted_ref import FormattedRef
from .localized_string import LocalizedString
from .typed_title_string import TypedTitleString


class SeriesType(str, Enum):
    MAIN = "main"
    ALT = "alt"
    JOURNAL = "journal"

    @classmethod
    def has_value(cls, value):
        return value in cls._value2member_map_


@dataclass(frozen=True)
class Series:
    type: str = None
    formattedref: FormattedRef = None
    title: TypedTitleString = None
    place: str = None
    organization: str = None
    abbreviation: LocalizedString = None
    from_: str = None
    to: str = None
    number: str = None
    partnumber: str = None

    def __post_init__(self):
        if not(type(self.title) is TypedTitleString or self.formattedref):
            raise ValueError("arg `title` or `formattedref` should present")

        if self.type and not SeriesType.has_value(self.type):
            logging.warning(
                f"[relaton-bib] Series type is invalid: {self.type}")

    # to_hash -> dataclasses.asdict

    def to_xml(self, parent, opts={}):
        name = "series"
        node = ET.Element(name) if parent is None \
            else ET.SubElement(parent, name)

        if self.formattedref:
            self.formattedref.to_xml(node)
        else:
            if self.title:
                self.title.to_xml(ET.SubElement(node, "title"))
            if self.place:
                ET.SubElement(node, "place").text = self.place
            if self.organization:
                ET.SubElement(node, "organization").text = self.organization
            if self.abbreviation:
                self.abbreviation.to_xml(ET.SubElement(node, "abbreviation"))
            if self.from_:
                ET.SubElement(node, "from").text = self.from_
            if self.to:
                ET.SubElement(node, "to").text = self.to
            if self.number:
                ET.SubElement(node, "number").text = self.number
            if self.partnumber:
                ET.SubElement(node, "partnumber").text = self.partnumber

        if self.type:
            node.attrib["type"] = self.type

        return node

    def to_asciibib(self, prefix="", count=1):
        pref = f"{prefix}.series" if prefix else "series"

        lines = [f"{pref}::"] if count > 1 else []
        if self.type:
            lines.append(f"{pref}.type:: {self.type}")
        if self.formattedref:
            lines.append(self.formattedref.to_asciibib(pref))
        if self.title:
            lines.append(self.title.to_asciibib(pref))
        if self.place:
            lines.append(f"{pref}.place:: {self.place}")
        if self.organization:
            lines.append(f"{pref}.organization:: {self.organization}")
        if self.abbreviation:
            lines.append(self.abbreviation.to_asciibib(f"{pref}.abbreviation"))
        if self.from_:
            lines.append(f"{pref}.from:: {self.from_}")
        if self.to:
            lines.append(f"{pref}.to:: {self.to}")
        if self.number:
            lines.append(f"{pref}.number:: {self.number}")
        if self.partnumber:
            lines.append(f"{pref}.partnumber:: {self.partnumber}")

        return "\n".join(lines)



import dataclasses
import inspect
import pytest

import xml.etree.ElementTree as ET

from relaton_bib import BibliographicItemVersion


@pytest.fixture
def subject():
    return BibliographicItemVersion(
        revision_date="01/01/2022",
        draft=["draft1", "2", "3"])


def test_to_xml(subject):
    host = ET.Element("host")
    result = subject.to_xml(host)

    assert len(result.findall("./draft")) == 3
    assert host.find("./version/revision-date").text == "01/01/2022"


def test_to_asciibib(subject):
    result = subject.to_asciibib()

    assert result == inspect.cleandoc(
        """version.revision_date:: 01/01/2022
           version.draft:: draft1
           version.draft:: 2
           version.draft:: 3""")


def test_to_asciibib_with_pref(subject):
    result = subject.to_asciibib(prefix="test")

    assert result == inspect.cleandoc(
        """test.version.revision_date:: 01/01/2022
           test.version.draft:: draft1
           test.version.draft:: 2
           test.version.draft:: 3""")


def test_hash(subject):
    result = dataclasses.asdict(subject)

    assert len(result["draft"]) == 3
    assert result["revision_date"] == "01/01/2022"



from dataclasses import dataclass

import xml.etree.ElementTree as ET

from .workgroup import WorkGroup


@dataclass
class TechnicalCommittee:
    workgroup: WorkGroup

    # to_hash -> dataclasses.asdict - different

    def to_xml(self, parent=None):
        name = "technical-committee"
        node = ET.Element(name)if parent is None \
            else ET.SubElement(parent, name)
        self.workgroup.to_xml(node)
        return node

    def to_asciibib(self, prefix="", count=1):
        name = "technical_committee"
        pref = f"{prefix}.{name}" if prefix else name
        out = [f"{pref}::"] if count > 1 else []
        out.append(self.workgroup.to_asciibib(pref))
        return "\n".join(out)



import dataclasses
import inspect
import pytest

import xml.etree.ElementTree as ET

from relaton_bib import TypedUri


@pytest.fixture
def subject():
    return TypedUri(type="type", content="https://metanorma.com")


def test_to_xml(subject):
    host = ET.Element("host")
    result = subject.to_xml(host)

    assert result.attrib["type"] == "type"
    assert host.find("./uri").text == "https://metanorma.com"


def test_to_asciibib(subject):
    result = subject.to_asciibib()

    assert result == inspect.cleandoc(
        """link.type:: type
           link.content:: https://metanorma.com""")


def test_to_asciibib_with_pref(subject):
    result = subject.to_asciibib(prefix="pref")

    assert result == inspect.cleandoc(
        """pref.link.type:: type
           pref.link.content:: https://metanorma.com""")


def test_hash(subject):
    result = dataclasses.asdict(subject)

    assert result["type"] == "type"
    assert result["content"] == "https://metanorma.com"


def test_invalid_uri():
    with pytest.raises(ValueError) as excinfo:
        TypedUri(type="type", content=":some_malformed_uri//")
    assert "Invalid content" in str(excinfo.value)


def test_set_content():
    uri = TypedUri(type="src", content=None)
    uri.content = "http://example.com"
    assert uri.content == "http://example.com"



from __future__ import annotations
from dataclasses import dataclass, field, InitVar
from enum import Enum
from typing import List, Union, Tuple

import re
import xml.etree.ElementTree as ET

from .formatted_string import FormattedString, FormattedStringFormat
from .localized_string import LocalizedString
from .relaton_bib import delegate, to_ds_instance


@dataclass
class TypedTitleString:
    class Type(str, Enum):
        MAIN = "main"
        TINTRO = "title-intro"
        TMAIN = "title-main"
        TPART = "title-part"

    type: str = None
    title: FormattedString = None
    content: InitVar[Union[str, List[LocalizedString]]] = field(default=None)
    language: InitVar[List[str]] = field(default=[])
    script: InitVar[List[str]] = field(default=list())
    format: InitVar[str] = field(
        default=FormattedStringFormat.TEXT_PLAIN.value)

    def __post_init__(self, content, language, script, format):
        if self.title is None and content is None:
            raise ValueError("Argument title or content should be passed")

        if not self.title:
            self.title = FormattedString(
                content=content,
                language=language,
                script=script,
                format=format)

    def __str__(self):
        return str(self.title)

    @classmethod
    def from_string(cls, title, lang=[], script=[]) \
            -> TypedTitleStringCollection:
        types = [cls.Type.TINTRO, cls.Type.TMAIN, cls.Type.TPART]
        ttls = cls.split_title(title)
        tts = [None if p is None else
               cls(type=types[i].value,
                   content=p,
                   language=lang,
                   script=script)
               for i, p in enumerate(ttls)]
        tts = list(filter(None, tts))
        ttls = list(filter(None, ttls))
        tts.append(TypedTitleString(type=cls.Type.MAIN,
                                    content=" - ".join(ttls),
                                    language=lang,
                                    script=script))
        return TypedTitleStringCollection(tts)

    @classmethod
    def split_title(cls, title) -> Tuple[str, str, str]:
        ttls = re.sub(r"\w\.Imp\s?\d+\u00A0:\u00A0", "", title).split(" - ")
        if len(ttls) < 2:
            return [None, str(ttls[0]), None]
        else:
            return cls.intro_or_part(ttls)

    @classmethod
    def intro_or_part(cls, ttls) -> Tuple[str, str, str]:
        if re.match(r"^(Part|Partie) \d+:", ttls[1]):
            return (None, ttls[0], " -- ".join(ttls[1:]))
        else:
            parts = ttls[2:]
            part = None
            if any(parts):
                part = " -- ".join(parts)
            return (ttls[0], ttls[1], part)

    def to_xml(self, parent) -> ET.Element:
        if self.type:
            parent.attrib["type"] = self.type
        self.title.to_xml(parent)
        return parent

    def to_asciibib(self, prefix="", count=1) -> str:
        pref = f"{prefix}." if prefix else prefix
        out = [f"{pref}title::"] if count > 1 else []
        if self.type:
            out.append(f"{pref}title.type:: {self.type}")
        out.append(self.title.to_asciibib(f"{pref}title", 1, bool(self.type)))
        return "\n".join(out)


@dataclass
@delegate("titles", "append", "extend", "insert", "remove", "pop", "clear",
          "__getitem__", "__len__", "__iter__", "__reversed__",
          "__contains__")
class TypedTitleStringCollection():
    titles: List[TypedTitleString]

    def __post_init__(self):
        self.titles = list(map(to_ds_instance(TypedTitleString), self.titles))

    def lang(self, lang=None):
        return self.__class__(filter(
            lambda t: lang in t.title.language, self.titles)) \
            if lang else self

    def delete_title_part(self):
        self.titles = list(
            filter(lambda t: t.type != TypedTitleString.Type.TPART,
                   self.titles))

    # @param init [Array, Hash]
    # @return [RelatonBib::TypedTitleStringCollection]
    # def reduce(init)
    #   self.class.new @array.reduce(init) { |m, t| yield m, t }
    # end

    # @param title [RelatonBib::TypedTitleString]
    # # @return [self]
    # def <<(title)
    #   titles << title
    #   self
    # end

    # # @param tcoll [RelatonBib::TypedTitleStringCollection]
    # # @return [RelatonBib::TypedTitleStringCollection]
    # def +(tcoll)
    #   TypedTitleStringCollection.new titles + tcoll.titles
    # end

    # @param opts [Hash]
    # @option opts [Nokogiri::XML::Builder] XML builder
    # @option opts [String, Symbol] :lang language
    def to_xml(self, parent, opts={}):
        lang = opts.get("lang")
        filtered = list(filter(
            lambda t: t.title.language and lang in t.title.language,
            self.titles))
        titles = filtered if any(filtered) else self.titles
        for t in titles:
            t.to_xml(ET.SubElement(parent, "title"))

        return parent



import dataclasses
import inspect
import pytest

import xml.etree.ElementTree as ET

from relaton_bib import WorkGroup


@pytest.fixture
def subj_min():
    return WorkGroup(name="value")


@pytest.fixture
def subj_full():
    return WorkGroup(name="value",
                     number=1,
                     type="type")


def test_min_to_xml(subj_min):
    host = ET.Element("host")
    result = subj_min.to_xml(host)

    assert result.text == "value"


def test_full_to_xml(subj_full):
    host = ET.Element("host")
    result = subj_full.to_xml(host)

    assert result.text == "value"
    assert result.attrib["type"] == "type"
    assert result.attrib["number"] == "1"


def test_min_to_asciibib(subj_min):
    result = subj_min.to_asciibib()

    assert result == "name:: value"


def test_full_to_asciibib(subj_full):
    result = subj_full.to_asciibib()

    assert result == inspect.cleandoc(
        """name:: value
           number:: 1
           type:: type""")


def test_min_to_asciibib_with_pref(subj_min):
    result = subj_min.to_asciibib(prefix="test")

    assert result == "test.name:: value"


def test_full_to_asciibib_with_pref(subj_full):
    result = subj_full.to_asciibib(prefix="test")

    assert result == inspect.cleandoc(
        """test.name:: value
           test.number:: 1
           test.type:: type""")


def test_hash(subj_full):
    result = dataclasses.asdict(subj_full)

    assert result["type"] == "type"
    assert result["number"] == 1
    assert result["name"] == "value"



"""Top-level package for relaton-bib."""

__author__ = """Aliaksandr Babrykovich"""
__email__ = 'abobrikovich@gmail.com'
__version__ = '0.1.0'

from .address import Address
from .affiliation import Affiliation
from .bib_item_locality import BibItemLocality, Locality, \
    BibItemLocalityType, SourceLocalityStack, SourceLocality, LocalityStack
from .biblio_note import BiblioNote, BiblioNoteCollection
from .biblio_version import BibliographicItemVersion
from .bibliographic_date import BibliographicDate, BibliographicDateType
from .bibliographic_item import BibliographicItem, BibliographicItemType
from .classification import Classification
from .contact import Contact, ContactType
from .contributor import Contributor
from .contribution_info import ContributionInfo, ContributorRole
from .copyright_association import CopyrightAssociation
from .document_identifier import DocumentIdentifier, DocumentIdType
from .document_relation_collection import DocRelationCollection
from .document_relation import DocumentRelation
from .document_status import DocumentStatus
from .editorial_group import EditorialGroup
from .formatted_ref import FormattedRef
from .formatted_string import FormattedString, FormattedStringFormat
from .hit import Hit
from .hit_collection import HitCollection
from .ics import ICS
from .organization import Organization, OrgIdentifier
from .localized_string import LocalizedString
from .medium import Medium
from .place import Place
from .person import Person, FullName, PersonIdentifier
from .series import Series, SeriesType
from .structured_identifier import StructuredIdentifier
from .structured_identifier import StructuredIdentifierCollection
from .typed_title_string import TypedTitleString, TypedTitleStringCollection
from .technical_committee import TechnicalCommittee
from .typed_uri import TypedUri
from .validity import Validity
from .workgroup import WorkGroup

from .bibtex_parser import from_bibtex
from .xml_parser import from_xml
from .dict_parser import from_dict

__all__ = [
    from_bibtex,
    from_xml,
    from_dict,
    BibliographicItem,
    BibliographicItemType,
    Address,
    Contact,
    ContactType,
    Contributor,
    Hit,
    HitCollection,
    Affiliation,
    TypedUri,
    DocumentIdentifier, DocumentIdType,
    CopyrightAssociation,
    FormattedString, FormattedStringFormat,
    ContributionInfo, ContributorRole,
    BibliographicDate, BibliographicDateType,
    Series, SeriesType,
    DocumentStatus,
    Organization, OrgIdentifier,
    LocalizedString,
    TypedTitleString, TypedTitleStringCollection,
    TechnicalCommittee,
    FormattedRef,
    Medium,
    Classification,
    Validity,
    BibItemLocality,
    Locality,
    BibItemLocalityType,
    SourceLocalityStack,
    SourceLocality,
    LocalityStack,
    BiblioNote,
    BiblioNoteCollection,
    BibliographicItemVersion,
    Place,
    Person, FullName, PersonIdentifier,
    StructuredIdentifierCollection,
    EditorialGroup,
    ICS,
    DocRelationCollection,
    DocumentRelation,
    WorkGroup,
    StructuredIdentifier
]



from __future__ import annotations
from dataclasses import dataclass, field
from typing import List

import xml.etree.ElementTree as ET

from .organization import Organization
from .localized_string import LocalizedString
from .formatted_string import FormattedString
from .relaton_bib import lang_filter


@dataclass
class Affiliation:
    organization: Organization
    name: LocalizedString = None
    description: List[FormattedString] = field(default_factory=list)

    def to_xml(self, parent, opts={}):
        name = "affiliation"
        result = ET.Element(name) if parent is None \
            else ET.SubElement(parent, name)
        if self.name:
            self.name.to_xml(ET.SubElement(result, "name"))

        for d in lang_filter(self.description, opts):
            d.to_xml(ET.SubElement(result, "description"))

        self.organization.to_xml(result, opts)

        return result

    def to_asciibib(self, prefix="", count=1):
        pref = f"{prefix}." if prefix else prefix
        out = f"#{pref}affiliation::" if count > 1 else []
        if self.name:
            out.append(self.name.to_asciibib(f"{pref}affiliation.name"))
        for d in self.description:
            out.append(d.to_asciibib(f"{pref}affiliation.description",
                                     len(self.description)))
        out.append(self.organization.to_asciibib(f"{pref}affiliation.*"))
        return "\n".join(out)



from __future__ import annotations
from dataclasses import dataclass, field
from enum import Enum
from typing import List

import re
import xml.etree.ElementTree as ET

from .relaton_bib import lang_filter, to_ds_instance
from .localized_string import LocalizedString
from .affiliation import Affiliation
from .contributor import Contributor


def localized_string():
    return to_ds_instance(LocalizedString)


@dataclass
class FullName:
    surname: LocalizedString = None
    completename: LocalizedString = None
    forename: List[LocalizedString] = field(default_factory=list)
    initial: List[LocalizedString] = field(default_factory=list)
    addition: List[LocalizedString] = field(default_factory=list)
    prefix: List[LocalizedString] = field(default_factory=list)

    def __post_init__(self):
        if not self.surname and not self.completename:
            raise ValueError("Should be given surname or completename")

        if isinstance(self.surname, str):
            self.surname = LocalizedString(self.surname)

        if isinstance(self.completename, str):
            self.completename = LocalizedString(self.completename)

        if isinstance(self.forename, str):
            self.forename = [LocalizedString(self.forename)]
        elif isinstance(self.forename, List):
            self.forename = list(map(localized_string(), self.forename))

        if isinstance(self.initial, str):
            self.initial = [LocalizedString(self.initial)]
        elif isinstance(self.initial, List):
            self.initial = list(map(localized_string(), self.initial))

        if isinstance(self.addition, str):
            self.addition = [LocalizedString(self.addition)]
        elif isinstance(self.addition, List):
            self.addition = list(map(localized_string(), self.addition))

        if isinstance(self.prefix, str):
            self.prefix = [LocalizedString(self.prefix)]
        elif isinstance(self.prefix, List):
            self.prefix = list(map(localized_string(), self.prefix))

    def to_xml(self, parent, opts={}):
        name = "name"
        result = ET.Element(name) if parent is None \
            else ET.SubElement(parent, name)

        if self.completename:
            self.completename.to_xml(ET.SubElement(result, "completename"))
        else:
            for p in lang_filter(self.prefix, opts):
                p.to_xml(ET.SubElement(result, "prefix"))
            for f in lang_filter(self.forename, opts):
                f.to_xml(ET.SubElement(result, "forename"))
            for i in lang_filter(self.initial, opts):
                i.to_xml(ET.SubElement(result, "initial"))
            self.surname.to_xml(ET.SubElement(result, "surname"))
            for a in lang_filter(self.addition, opts):
                a.to_xml(ET.SubElement(result, "addition"))

        return result

    def to_asciibib(self, prefix):
        prf = f"{prefix}.name." if prefix else "name."
        out = [fn.to_asciibib(f"{prf}forename", len(self.forename))
               for fn in self.forename]
        out += [i.to_asciibib(f"{prf}initial", len(self.initial))
                for i in self.initial]
        if self.surname:
            out.append(self.surname.to_asciibib(f"{prf}surname"))
        out += [ad.to_asciibib(f"{prf}addition", len(self.addition))
                for ad in self.addition]
        out += [pr.to_asciibib(f"{prf}prefix", len(self.prefix))
                for pr in self.prefix]
        if self.completename:
            out.append(self.completename.to_asciibib(f"{prf}completename"))
        return "\n".join(out)


class PersonIdentifierType(Enum):
    ISNI = "isni"
    URI = "uri"

    @classmethod
    def has_value(cls, value):
        return value in cls._value2member_map_


@dataclass
class PersonIdentifier:
    type: str
    value: str

    def __post_init__(self):
        if not PersonIdentifierType.has_value(self.type):
            raise ValueError('Invalid type. It should be "isni" or "uri".')

        if isinstance(self.type, PersonIdentifierType):
            self.type = self.type.value

    def to_xml(self, parent):
        name = "identifier"
        result = ET.Element(name) if parent is None \
            else ET.SubElement(parent, name)
        result.text = self.value
        result.attrib["type"] = self.type

    def to_asciibib(self, prefix="", count=1):
        pref = prefix + "." if prefix else prefix
        out = [f"{prefix}::"] if count > 1 else []
        out.append(f"{pref}type:: {self.type}")
        out.append(f"{pref}value:: {self.value}")
        return "\n".join(out)


@dataclass
class Person(Contributor):
    name: FullName = None
    affiliation: List[Affiliation] = field(default_factory=list)
    identifier: List[PersonIdentifier] = field(default_factory=list)

    def __post_init__(self):
        # WORKAROUND for https://bugs.python.org/issue36077
        if not self.name:
            raise ValueError("missing name")

    def to_xml(self, parent, opts={}):
        name = "person"
        result = ET.Element(name) if parent is None \
            else ET.SubElement(parent, name)
        self.name.to_xml(result, opts)
        for a in self.affiliation:
            a.to_xml(result, opts)
        for i in self.identifier:
            i.to_xml(result)
        for c in self.contact:
            c.to_xml(result)
        return result

    def to_asciibib(self, prefix="", count=1):
        pref = re.sub(r"\*$", "person", prefix)
        out = [f"{pref}::"] if count > 1 else []
        out.append(self.name.to_asciibib(pref))
        for a in self.affiliation:
            out.append(a.to_asciibib(pref, len(self.affiliation)))
        for i in self.identifier:
            out.append(i.to_asciibib(pref, len(self.identifier)))
        out.append(super().to_asciibib(pref))
        return "\n".join([l for l in out if l])

    def bib_name(self) -> str:
        return str(self.name.completename)


from dataclasses import dataclass
from urllib.parse import urlparse

import xml.etree.ElementTree as ET


@dataclass()
class TypedUri:
    content: str
    type: str = None

    def __post_init__(self):
        if not self._valid_uri:
            raise ValueError(f"Invalid content: {self.content}")

    # to_hash -> dataclasses.asdict

    def to_xml(self, parent=None, opts={}):
        name = "uri"
        node = ET.Element(name) if parent is None \
            else ET.SubElement(parent, name)
        node.text = self.content

        if self.type:
            node.attrib["type"] = self.type

        return node

    def to_asciibib(self, prefix="", count=1):
        pref = f"{prefix}.link" if prefix else "link"
        out = [f"{pref}::"] if count > 1 else []
        if self.type:
            out.append(f"{pref}.type:: {self.type}")
        out.append(f"{pref}.content:: {self.content}")
        return "\n".join(out)

    @property
    def _valid_uri(self):
        if self.content is None:
            return True
        try:
            result = urlparse(self.content)
            return all([result.scheme, result.netloc])
        except Exception:
            return False



import xml.etree.ElementTree as ET

from dataclasses import dataclass, InitVar
from typing import List

from .formatted_string import FormattedString
from .relaton_bib import delegate


@dataclass(frozen=True)
class BiblioNote(FormattedString):
    type: str = None
    format: str = None

    def to_xml(self, parent=None):
        name = "note"
        node = ET.Element(name) if parent is None \
            else ET.SubElement(parent, name)
        super().to_xml(node)
        if self.type:
            node.attrib["type"] = self.type
        return node

    def to_asciibib(self, prefix="", count=1):
        pref = f"{prefix}." if prefix else prefix
        has_attrs = self.type
        out = [f"{pref}biblionote::"] if count > 1 and has_attrs else []
        if self.type:
            out.append(f"{pref}biblionote.type:: {self.type}")
        out.append(super().to_asciibib(f"{pref}biblionote", 1, has_attrs))
        return "\n".join(out)


@dataclass
@delegate("array", "append", "__getitem__", "__len__", "__iter__",
          "__reversed__", "__contains__")
class BiblioNoteCollection:
    array: List[BiblioNote]

    @property
    def first(self):
        return self.array[0] if len(self.array) > 0 else None

    @property
    def last(self):
        return self.array[-1] if len(self.array) > 0 else None

    @property
    def empty(self):
        return len(self.array) == 0

    def any(self):
        # https://stackoverflow.com/a/2323165/902217
        return any(x for x in self.array)

    def to_xml(self, parent, opts={}):
        lang = opts.get("lang")
        bnc = list(filter(lambda bn: lang in bn.language, self.array))
        if not bnc:
            bnc = self.array
        [bn.to_xml(parent) for bn in bnc]
        return parent

    # TODO missing to_asciibib?
    # def to_asciibib(self, prefix="", count=1):
    #   return "\n".join(map(lambda bn: bn.to_asciibib(prefix, count), self.array))



from __future__ import annotations
from dataclasses import dataclass

import xml.etree.ElementTree as ET

from .relaton_bib import to_ds_instance


@dataclass
class DocumentStatus:

    @dataclass
    class Stage:
        value: str
        abbreviation: str = None

        def to_xml(self, parent):
            if self.abbreviation:
                parent.attrib["abbreviation"] = self.abbreviation
            parent.text = self.value
            return parent

        # NOTE missing to_asciibib in original

    stage: Stage
    substage: Stage = None
    iteration: str = None

    def __post_init__(self):
        self.stage = to_ds_instance(DocumentStatus.Stage)(self.stage)
        self.substage = to_ds_instance(DocumentStatus.Stage)(self.substage)

    def to_xml(self, parent):
        name = "status"
        result = ET.Element(name) if parent is None \
            else ET.SubElement(parent, name)
        self.stage.to_xml(ET.SubElement(result, "stage"))
        if self.substage:
            self.substage.to_xml(ET.SubElement(result, "substage"))
        if self.iteration:
            ET.SubElement(result, "iteration").text = self.iteration

        return result

    def to_asciibib(self, prefix=""):
        pref = f"{prefix}." if prefix else prefix
        out = [f"{pref}docstatus.stage:: {self.stage.value}"]
        if self.substage:
            out.append(f"{pref}docstatus.substage:: {self.substage.value}")
        if self.iteration:
            out.append(f"{pref}docstatus.iteration:: {self.iteration}")
        return "\n".join(out)



from __future__ import annotations
from dataclasses import dataclass, field
from enum import Enum
from typing import Union

import logging
import xml.etree.ElementTree as ET

from .bibliographic_item import *
from .formatted_string import FormattedString
from .bib_item_locality import Locality, LocalityStack, SourceLocality, \
    SourceLocalityStack


@dataclass
class DocumentRelation:
    class Type(str, Enum):
        includes = "includes"
        includedIn = "includedIn"
        hasPart = "hasPart"
        partOf = "partOf"
        merges = "merges"
        mergedInto = "mergedInto"
        splits = "splits"
        splitInto = "splitInto"
        instance = "instance"
        hasInstance = "hasInstance"
        exemplarOf = "exemplarOf"
        hasExemplar = "hasExemplar"
        manifestationOf = "manifestationOf"
        hasManifestation = "hasManifestation"
        reproductionOf = "reproductionOf"
        hasReproduction = "hasReproduction"
        reprintOf = "reprintOf"
        hasReprint = "hasReprint"
        expressionOf = "expressionOf"
        hasExpression = "hasExpression"
        translatedFrom = "translatedFrom"
        hasTranslation = "hasTranslation"
        arrangementOf = "arrangementOf"
        hasArrangement = "hasArrangement"
        abridgementOf = "abridgementOf"
        hasAbridgement = "hasAbridgement"
        annotationOf = "annotationOf"
        hasAnnotation = "hasAnnotation"
        draftOf = "draftOf"
        hasDraft = "hasDraft"
        editionOf = "editionOf"
        hasEdition = "hasEdition"
        updates = "updates"
        updatedBy = "updatedBy"
        derivedFrom = "derivedFrom"
        derives = "derives"
        describes = "describes"
        describedBy = "describedBy"
        catalogues = "catalogues"
        cataloguedBy = "cataloguedBy"
        hasSuccessor = "hasSuccessor"
        successorOf = "successorOf"
        adaptedFrom = "adaptedFrom"
        hasAdaptation = "hasAdaptation"
        adoptedFrom = "adoptedFrom"
        adoptedAs = "adoptedAs"
        reviewOf = "reviewOf"
        hasReview = "hasReview"
        commentaryOf = "commentaryOf"
        hasCommentary = "hasCommentary"
        related = "related"
        complements = "complements"
        complementOf = "complementOf"
        obsoletes = "obsoletes"
        obsoletedBy = "obsoletedBy"
        cited = "cited"
        isCitedIn = "isCitedIn"
        replace = "replace"  # special type to be replaced

        @classmethod
        def has_value(cls, value):
            return value in cls._value2member_map_

    type: str
    bibitem: BibliographicItem
    description: FormattedString = None
    locality: Union[Locality, LocalityStack] = field(default_factory=list)
    source_locality: Union[SourceLocality, SourceLocalityStack] = field(
        default_factory=list)

    def __post_init__(self):
        if self.type == "Now withdrawn":
            self.type = DocumentRelation.Type.obsoletes

        if not DocumentRelation.Type.has_value(self.type):
            logging.warning(
                f"[relaton-bib] WARNING: invalid relation type: {self.type}")

        if isinstance(self.type, DocumentRelation.Type):
            self.type = self.type.value

    def to_xml(self, parent, opts={}) -> str:
        opts.pop("bibdata", None)
        opts.pop("note", None)

        name = "relation"
        result = ET.Element(name) if parent is None \
            else ET.SubElement(parent, name)
        result.attrib["type"] = self.type
        if self.description:
            self.description.to_xml(ET.SubElement(result, "description"))

        self.bibitem.to_xml(result, {"embedded": True, **opts})

        for loc in self.locality:
            loc.to_xml(result)

        for loc in self.source_locality:
            loc.to_xml(result)

        return result

    def to_asciibib(self, prefix="", count=1) -> str:
        pref = f"{prefix}." if prefix else prefix
        out = [f"{pref}type:: {self.type}"]
        if self.description:
            out.append(self.description.to_asciibib(f"{pref}desctiption"))
        if self.bibitem:
            out.append(self.bibitem.to_asciibib(f"{pref}bibitem"))
        return "\n".join(out)



import dataclasses
import datetime
import inspect
import pytest

import xml.etree.ElementTree as ET

from relaton_bib import CopyrightAssociation, ContributionInfo, Organization


@pytest.fixture
def subject():
    return CopyrightAssociation(
        from_=datetime.date(1990, 7, 26),
        owner=[
            ContributionInfo(role=[], entity=Organization(name="org1")),
            ContributionInfo(role=[], entity=Organization(name="org2"))
        ],
        scope="scope")


@pytest.fixture
def subject2():
    return CopyrightAssociation(from_="1991", owner=[
        ContributionInfo(role=[], entity=Organization(name="org3")),
        ContributionInfo(role=[], entity=Organization(name="org4"))]
    )


def test_error_empty_owners():
    with pytest.raises(ValueError) as excinfo:
        CopyrightAssociation(from_="2019", owner=[])
    assert "at least one owner should exist." \
        in str(excinfo.value)


def test_to_xml(subject):
    host = ET.Element("host")
    result = subject.to_xml(host)

    assert host.find("./copyright/from").text == "1990"
    assert len(host.findall("./copyright/owner")) == 2
    assert host.find("./copyright/scope").text == "scope"


def test_to_xml_2(subject2):
    host = ET.Element("host")
    result = subject2.to_xml(host)

    assert host.find("./copyright/from").text == "1991"
    assert len(host.findall("./copyright/owner/organization")) == 2
    assert host.find("./copyright/scope") is None


def test_to_asciibib(subject):
    result = subject.to_asciibib()

    assert result == inspect.cleandoc(
        """copyright::
           copyright.owner.name:: org1
           copyright.role.type:: publisher
           copyright::
           copyright.owner.name:: org2
           copyright.role.type:: publisher
           copyright.from:: 1990
           copyright.scope:: scope""")


def test_to_asciibib_with_pref(subject):
    result = subject.to_asciibib(prefix="tst")

    # role have no copyright because of logic in contribution_info.py:87
    assert result == inspect.cleandoc(
        """tst::
           tst.copyright.owner.name:: org1
           tst.role.type:: publisher
           tst::
           tst.copyright.owner.name:: org2
           tst.role.type:: publisher
           tst.copyright.from:: 1990
           tst.copyright.scope:: scope""")


def test_hash(subject):
    result = dataclasses.asdict(subject)

    assert type(result["from_"]) is datetime.date
    assert result["scope"] == "scope"



from dataclasses import dataclass

import xml.etree.ElementTree as ET


@dataclass(frozen=True)
class Medium:
    form: str = None
    size: str = None
    scale: str = None

    def to_xml(self, parent):
        name = "medium"
        node = ET.Element(name) if parent is None \
            else ET.SubElement(parent, name)
        if self.form:
            ET.SubElement(node, "form").text = self.form
        if self.size:
            ET.SubElement(node, "size").text = self.size
        if self.scale:
            ET.SubElement(node, "scale").text = self.scale
        return node

    def to_asciibib(self, prefix=""):
        pref = f"{prefix}.medium." if prefix else "medium."
        out = []
        if self.form:
            out.append(f"{pref}form:: {self.form}")
        if self.size:
            out.append(f"{pref}size:: {self.size}")
        if self.scale:
            out.append(f"{pref}scale:: {self.scale}")
        return "\n".join(out)



import dataclasses
import inspect
import pytest

import xml.etree.ElementTree as ET

from relaton_bib import Medium


@pytest.fixture
def subject():
    return Medium(form="form", size="size", scale="scale")


@pytest.fixture
def subject_none():
    return Medium()


def test_to_xml(subject):
    host = ET.Element("host")
    result = subject.to_xml(host)

    assert host.find("./medium/form").text == "form"
    assert host.find("./medium/size").text == "size"
    assert result.find("./scale").text == "scale"


def test_to_xml_with_no_props(subject_none):
    host = ET.Element("host")
    result = subject_none.to_xml(host)

    assert result is not None
    assert host.find("./medium") is not None


def test_to_asciibib(subject):
    result = subject.to_asciibib()

    assert result == inspect.cleandoc("""medium.form:: form
                                         medium.size:: size
                                         medium.scale:: scale""")


def test_to_asciibib_with_pref(subject):
    result = subject.to_asciibib("p")

    assert result == inspect.cleandoc("""p.medium.form:: form
                                         p.medium.size:: size
                                         p.medium.scale:: scale""")


def test_hash(subject):
    result = dataclasses.asdict(subject)

    assert result["form"] == "form"
    assert result["size"] == "size"
    assert result["scale"] == "scale"



from __future__ import annotations
from dataclasses import dataclass
from typing import List

from .document_relation import DocumentRelation
from .relaton_bib import delegate


@dataclass
@delegate("array", "append", "__getitem__", "__len__", "__iter__",
          "__reversed__", "__contains__")
class DocRelationCollection:
    array: List[DocumentRelation]

    def __post_init__(self):
        self.array = list(map(
            lambda r: DocumentRelation(**r) if isinstance(r, dict) else r,
            self.array))

    @property
    def first(self):
        return self.array[0] if len(self.array) > 0 else None

    @property
    def last(self):
        return self.array[-1] if len(self.array) > 0 else None

    @property
    def empty(self):
        return len(self.array) == 0

    def any(self):
        # https://stackoverflow.com/a/2323165/902217
        return any(x for x in self.array)

    def to_xml(self, parent, opts={}):
        # NOTE original gem missing to_xml
        for r in self.array:
            r.to_xml(parent, opts)
        return parent

    def replaces(self):
        return DocRelationCollection(filter(
            lambda r: r.type == DocumentRelation.Type.replace, self.array))

    def to_asciibib(self, prefix=""):
        pref = f"{prefix}.relation" if prefix else "relation"
        out = []
        for r in self.array:
            if len(self) > 1:
                out.append(f"{pref}::")
            out.append(r.to_asciibib(pref))
        return "\n".join(out)



import re
import datetime
import logging
import xml.etree.ElementTree as ET

from dataclasses import dataclass
from enum import Enum
from typing import ClassVar

from .relaton_bib import parse_date


class BibliographicDateType(str, Enum):
    PUBLISHED = "published"
    ACCESSED = "accessed"
    CREATED = "created"
    IMPLEMENTED = "implemented"
    OBSOLETED = "obsoleted"
    CONFIRMED = "confirmed"
    UPDATED = "updated"
    ISSUED = "issued"
    TRANSMITTED = "transmitted"
    COPIED = "copied"
    UNCHANGED = "unchanged"
    CIRCULATED = "circulated"
    ADAPTED = "adapted"
    VOTE_STARTE = "vote-started"
    VOTE_ENDED = "vote-ended"

    @classmethod
    def has_value(cls, value):
        return value in cls._value2member_map_


@dataclass
class BibliographicDate:
    NO_YEAR: ClassVar[str] = "--"

    type: str
    on: datetime.date = None
    from_: datetime.date = None
    to: datetime.date = None

    def __post_init__(self):
        if not BibliographicDateType.has_value(self.type):
            logging.warning(
                f"[relaton-bib] invalid bibliographic date type: {self.type}")

        if isinstance(self.type, BibliographicDateType):
            self.type = self.type.value

        if not (self.on or self.from_):
            raise ValueError("expected on or from_ argument")

        if self.on and isinstance(self.on, str):
            self.on = parse_date(self.on)

        if self.from_ and isinstance(self.from_, str):
            self.from_ = parse_date(self.from_)

        if self.to and isinstance(self.to, str):
            self.to = parse_date(self.to)

    def to_xml(self, parent, opts={}):
        name = "date"
        result = ET.Element(name) if parent is None \
            else ET.SubElement(parent, name)
        result.attrib["type"] = self.type
        date_format = opts.get("date_format")
        no_year = opts.get("no_year")
        if self.on:
            on_node = ET.SubElement(result, "on")
            if no_year:
                on_node.text = self.NO_YEAR
            else:
                on_node.text = self.date_format(self.on, date_format)
        elif self.from_:
            form_node = ET.SubElement(result, "from")
            if no_year:
                form_node.text = self.NO_YEAR
            else:
                form_node.text = self.date_format(self.from_, date_format)
                if self.to:
                    to_node = ET.SubElement(result, "to")
                    to_node.text = self.date_format(self.to, date_format)

        return result

    def to_asciibib(self, prefix="", count=1):
        pref = prefix + "." if prefix else prefix
        out = [f"{pref}date::"] if count > 1 else []
        out.append(f"{pref}date.type:: {self.type}")
        if self.on:
            out.append(f"{pref}date.on:: {self.on}")
        if self.from_:
            out.append(f"{pref}date.from:: {self.from_}")
        if self.to:
            out.append(f"{pref}date.to:: #{self.to}")
        return "\n".join(out)

    def value(self, prop="on", part=None) -> str:
        return self._process_date(getattr(self, prop), part)

    def part(self, part="year") -> str:
        return self.value("on", part) or self.value("from", part) \
            or self.value("to", part)

    def date_format(self, date, fmt_name=None):
        if fmt_name == "short":
            fmt = "%Y-%m"
        elif fmt_name == "full":
            fmt = "%Y-%m-%d"
        else:
            return date

        date = self._parse_date(date)

        return date.strftime(fmt) if isinstance(date, datetime.date) else date

    def _process_date(self, date, part=None):
        if not (date and part):
            return date

        date = self._parse_date(date)

        if part == "date":
            return date

        return getattr(date, part) if isinstance(date, datetime.date) else date

    def _parse_date(self, date):
        if re.match(r"^\d{4}-\d{2}-\d{2}", date):
            return datetime.datetime.strptime(date, "%Y-%m-%d")  # 2012-02-11
        elif re.match(r"^\d{4}-\d{2}", date):
            return datetime.datetime.strptime(date, "%Y-%m")  # 2012-02
        elif re.match(r"^\d{4}", date):
            return datetime.datetime.strptime(date, "%Y")  # 2012
        else:
            return date



from __future__ import annotations
from dataclasses import dataclass, field
from typing import List, Dict

from .hit_collection import HitCollection


@dataclass
class Hit:
    hit: List[Dict] = field(default_factory=list)
    hit_collection: HitCollection = None

    def fetch():
        raise NotImplementedError()

    # @param opts [Hash]
    # @option opts [Nokogiri::XML::Builder] :builder XML builder
    # @option opts [Boolean] :bibdata
    # @option opts [String, Symbol] :lang language
    # @return [String] XML
    def to_xml(self, parent, opts={}):
        self.fetch().to_xml(parent, opts)



from dataclasses import dataclass
from typing import Optional

import xml.etree.ElementTree as ET


@dataclass(frozen=True)
class Classification:
    value: str
    type: str = None

    def to_xml(self, parent):
        name = "classification"
        node = ET.Element(name) if parent is None \
            else ET.SubElement(parent, name)
        node.text = self.value

        if self.type:
            node.attrib["type"] = self.type

        return node

    def to_asciibib(self, prefix="", count=1):
        pref = f"{prefix}.classification" if prefix else "classification"
        out = [f"{pref}::"] if count > 1 else []
        if self.type:
            out.append(f"{pref}.type:: {self.type}")
        out.append(f"{pref}.value:: {self.value}")
        return "\n".join(out)



import dataclasses
import inspect
import pytest

import xml.etree.ElementTree as ET

from relaton_bib import DocumentIdentifier, DocumentIdType


@pytest.fixture
def subj_iso():
    return DocumentIdentifier(
        id="1111-2:2014",
        type=DocumentIdType.ISO)


def test_iso_remove_part(subj_iso):
    subj_iso.remove_part()
    assert subj_iso.id == "1111:2014"
    subj_iso.all_parts()
    assert subj_iso.id == "1111:2014 (all parts)"


def test_iso_remove_date(subj_iso):
    subj_iso.remove_date()
    subj_iso.id == "1111-2"


@pytest.fixture
def subj_urn_iso():
    return DocumentIdentifier(
        id="urn:iso:std:iso:1111:-1:stage-60.60:ed-1:v1:en,fr",
        type=DocumentIdType.URN)


def test_urn_iso_remove_part(subj_urn_iso):
    subj_urn_iso.remove_part()
    subj_urn_iso.id == "urn:iso:std:iso:1111"


@pytest.fixture
def subj_urn_iec():
    return DocumentIdentifier(
        type=DocumentIdType.URN,
        id="urn:iec:std:iec:61058-2-4:1995::csv:en:plus:amd:1:2003")


def test_urn_iec_remove_part(subj_urn_iec):
    subj_urn_iec.remove_part()
    subj_urn_iec.id = "urn:iec:std:iec:61058:1995::csv:en:plus:amd:1:2003"


def test_urn_iec_remove_date(subj_urn_iec):
    subj_urn_iec.remove_date()
    subj_urn_iec.id == "urn:iec:std:iec:61058-2-4:::csv:en:plus:amd:1:2003"


def test_urn_iec_set_all_parts(subj_urn_iec):
    subj_urn_iec.all_parts()
    subj_urn_iec.id == "urn:iec:std:iec:61058-2-4:1995::ser"


@pytest.fixture
def subj_gb():
    return DocumentIdentifier(
        id="1111.2-2014",
        type="Chinese Standard")  # test string not enum


def test_gb_remove_part(subj_gb):
    subj_gb.remove_part()
    subj_gb.id == "1111-2014"


def test_gb_remove_date(subj_gb):
    subj_gb.remove_date()
    subj_gb.id == "1111.2"


def test_to_xml(subj_iso):
    host = ET.Element("host")
    result = subj_iso.to_xml(host)

    assert result.attrib["type"] == "ISO"
    assert result.text == "1111-2:2014"


def test_to_asciibib(subj_urn_iso):
    result = subj_urn_iso.to_asciibib()

    assert result == inspect.cleandoc(
        """docid.type:: URN
           docid.id:: urn:iso:std:iso:1111:-1:stage-60.60:ed-1:v1:en,fr""")


def test_to_asciibib_with_pref(subj_urn_iec):
    result = subj_urn_iec.to_asciibib(prefix="test")

    assert result == inspect.cleandoc("""
        test.docid.type:: URN
        test.docid.id:: urn:iec:std:iec:61058-2-4:1995::csv:en:plus:amd:1:2003
        """)


def test_hash(subj_gb):
    result = dataclasses.asdict(subj_gb)

    assert result["type"] == "Chinese Standard"
    assert result["scope"] is None



import datetime
import typing

import xml.etree.ElementTree as ET

from dataclasses import dataclass


@dataclass(frozen=True)
class Validity:
    FORMAT: typing.ClassVar[str] = "%Y-%m-%d %H:%M"

    begins: datetime.datetime = None
    ends: datetime.datetime = None
    revision: datetime.datetime = None

    # to_hash -> dataclasses.asdict TODO original convert datetime -> str

    def to_xml(self, parent):
        name = "validity"
        result = ET.Element(name) if parent is None \
            else ET.SubElement(parent, name)

        if self.begins:
            ET.SubElement(result, "validityBegins").text = self.begins.strftime(self.FORMAT)
        if self.ends:
            ET.SubElement(result, "validityEnds").text = self.ends.strftime(self.FORMAT)
        if self.revision:
            ET.SubElement(result, "revision").text = self.revision.strftime(self.FORMAT)

        return result

    # @param prefix [String]
    # @return [String]
    def to_asciibib(self, prefix=""):
        """Return AsciiBib representation

        Keyword arguments:
        prefix -- AsciiBib prefix
        """
        pref = f"{prefix}.validity." if prefix else "validity."
        out = []
        if self.begins:
            out.append(f"{pref}begins:: {self.begins.strftime(self.FORMAT)}")
        if self.ends:
            out.append(f"{pref}ends:: {self.ends.strftime(self.FORMAT)}")
        if self.revision:
            out.append(
                f"{pref}revision:: {self.revision.strftime(self.FORMAT)}")
        return "\n".join(out)



from __future__ import annotations
from dataclasses import dataclass, field
from enum import Enum
from typing import List
import re
import logging
import xml.etree.ElementTree as ET

from .localized_string import LocalizedString
from .relaton_bib import lang_filter, to_ds_instance
from .contributor import Contributor


class OrgIdentifierType(Enum):
    ORCID = "orcid"
    URI = "uri"

    @classmethod
    def has_value(cls, value):
        return value in cls._value2member_map_


@dataclass(frozen=True)
class OrgIdentifier:
    type: str
    value: str

    def __post_init__(self):
        if not (OrgIdentifierType.has_value(self.type)):
            logging.warning(
                f"[relaton-bib] invalid locality type: {self.type}")

    def to_xml(self, parent):
        name = "identifier"
        node = ET.Element(name) if parent is None \
            else ET.SubElement(parent, name)
        node.text = self.value
        node.attrib["type"] = self.type
        return node

    def to_asciibib(self, prefix="", count=1):
        pref = f"{prefix}." if prefix else prefix
        out = [f"{pref}identifier::"] if count > 1 else []
        out.append(f"{pref}identifier.type:: {self.type}")
        out.append(f"{pref}identifier.value:: {self.value}")
        return "\n".join(out)


@dataclass
class Organization(Contributor):
    name: List[LocalizedString] = field(default_factory=list)
    abbreviation: LocalizedString = None
    subdivision: List[LocalizedString] = field(default_factory=list)
    identifier: List[OrgIdentifier] = field(default_factory=list)

    def __post_init__(self):
        if not self.name:
            # WORKAROUND for https://bugs.python.org/issue36077
            raise ValueError("missing name")

        if isinstance(self.name, str):
            self.name = [LocalizedString(self.name)]
        elif isinstance(self.name, List):
            self.name = list(map(to_ds_instance(LocalizedString), self.name))

        if isinstance(self.abbreviation, str):
            self.abbreviation = LocalizedString(self.abbreviation)

        if isinstance(self.subdivision, List):
            self.subdivision = list(map(to_ds_instance(LocalizedString),
                                        self.subdivision))

    def to_xml(self, parent, opts={}):
        name = "organization"
        result = ET.Element(name) if parent is None \
            else ET.SubElement(parent, name)
        for n in lang_filter(self.name, opts):
            n.to_xml(ET.SubElement(result, "name"))
        for s in lang_filter(self.subdivision, opts):
            s.to_xml(ET.SubElement(result, "subdivision"))
        if self.abbreviation:
            self.abbreviation.to_xml(ET.SubElement(result, "abbreviation"))
        if self.uri:
            ET.SubElement(result, "uri").text = self.uri
        for idntfr in self.identifier:
            idntfr.to_xml(result)
        super().to_xml(result)
        return result

    def to_asciibib(self, prefix="", count=1):
        prefix = re.sub(r"\*$", "organization", prefix)
        out = [f"{prefix}::"] if count > 1 else []
        pref = f"{prefix}." if prefix else prefix
        for n in self.name:
            out.append(n.to_asciibib(f"{pref}name", len(self.name)))
        if self.abbreviation:
            out.append(self.abbreviation.to_asciibib(f"{pref}abbreviation"))
        for sd in self.subdivision:
            if len(self.subdivision) > 1:
                # NOTE originally it was without \n
                out.append(f"{pref}subdivision::")
            out.append(sd.to_asciibib(f"{pref}subdivision"))
        for idtfr in self.identifier:
            out.append(idtfr.to_asciibib(prefix, len(self.identifier)))
        parent = super().to_asciibib(prefix)
        if parent:
            out.append(parent)
        return "\n".join(out)

    def bib_name(self) -> str:
        return str(self.name[0]) if len(self.name) > 0 else None



from dataclasses import dataclass

import xml.etree.ElementTree as ET


@dataclass
class ICS:
    code: str
    text: str

    def to_xml(self, parent):
        name = "ics"
        node = ET.Element(name) if parent is None \
            else ET.SubElement(parent, name)
        ET.SubElement(node, "code").text = self.code
        ET.SubElement(node, "text").text = self.text
        return node

    def to_asciibib(self, prefix="", count=1):
        suffix = "ics"
        pref = f"{prefix}.{suffix}" if prefix else suffix
        out = [f"{pref}::"] if count > 1 else []
        out.append(f"{pref}.code:: {self.code}")
        out.append(f"{pref}.text:: {self.text}")
        return "\n".join(out)



import pytest

from relaton_bib import Hit, HitCollection


@pytest.fixture(autouse=True)
def mock_pool_map(monkeypatch):
    monkeypatch.setattr("multiprocessing.pool.Pool.map",
                        lambda self, f, it: [f(o) for o in it])


@pytest.fixture
def bibitem(mocker):
    item = mocker.Mock(name="bibitem")
    item.to_xml = mocker.MagicMock(return_value=None)
    return item


@pytest.fixture
def hit(mocker, bibitem):
    hit = Hit()
    hit.fetch = mocker.MagicMock(return_value=bibitem)
    return hit


@pytest.fixture
def subject(mocker, hit):
    hits = HitCollection("ref")
    hits.append(hit)
    return hits


def test_fetches_all_hits(subject, hit):
    subject.fetch()
    hit.fetch.assert_called_once()


def test_collection_to_xml(subject, mocker, hit, bibitem):
    result = subject.to_xml()

    hit.fetch.assert_has_calls([
       [mocker.call(), mocker.call()]
    ])
    bibitem.to_xml.assert_called_once()

    assert result is not None
    assert result.tag == "documents"


@pytest.mark.skip("skippend because there is different appoach for iteration")
def test_select_hits(subject):
    assert isinstance(filter(subject), HitCollection)


@pytest.mark.skip("skippend because there is different appoach for iteration")
def test_reduce_collection():
    import functools
    functools.reduce(lambda sum, hit: sum.append(hit), subject)
    isinstance(subject, HitCollection)



from __future__ import annotations
from dataclasses import dataclass
from enum import Enum

import logging
import re
import xml.etree.ElementTree as ET


class DocumentIdType(str, Enum):
    CN_STD = "Chinese Standard"
    ISO = "ISO"
    IEC = "IEC"
    URN = "URN"


@dataclass
class DocumentIdentifier:
    id: str
    type: str = None
    scope: str = None

    def remove_part(self):
        if self.type == DocumentIdType.CN_STD:
            self.id = re.sub(r"\.\d+", "", self.id)
        elif self.type in (DocumentIdType.ISO, DocumentIdType.IEC):
            self.id = re.sub(r"-[^:]+", "", self.id)
        elif self.type == DocumentIdType.URN:
            self._remove_urn_part()
        else:
            # sanity check
            logging.warning(f"[relaton-bib] unknown doc type: {self.type}")

    def remove_date(self):
        if self.type == DocumentIdType.CN_STD:
            self.id = re.sub(r"-[12]\d\d\d", "", self.id)
        elif self.type in (DocumentIdType.ISO, DocumentIdType.IEC):
            self.id = re.sub(r":[12]\d\d\d", "", self.id)
        elif self.type == DocumentIdType.URN:
            self.id = re.sub(
                r"^(urn:iec:std:[^:]+:[^:]+:)[^:]*", r"\1", self.id)

        # sanity check
        logging.warning(f"[relaton-bib] unknown doc type: {self.type}")

    def all_parts(self):
        if self.type == DocumentIdType.URN:
            return re.sub(r"^(urn:iec:std(?::[^:]*){4}).*", r"\1:ser", self.id)
        else:
            self.id += " (all parts)"

    def to_xml(self, parent, opts={}):
        lang = opts.get("lang")
        lid_re = fr"(?<=:)(?:\w\w,)*?({lang})(?:,\w\w)*"
        lid = re.sub(lid_re, r"\1", self.id) \
            if self.type == DocumentIdType.URN and lang else self.id

        result = ET.SubElement(parent, "docidentifier")
        result.text = lid
        if self.type:
            result.attrib["type"] = self.type
        if self.scope:
            result.attrib["scope"] = self.scope

        return result

    def to_asciibib(self, prefix="", count=1) -> str:
        pref = f"{prefix}." if prefix else prefix

        if not self.type and not self.scope:
            return f"{pref}docid:: {self.id}"

        out = [f"{pref}docid::"] if count > 1 else []
        if self.type:
            out.append(f"{pref}docid.type:: {self.type}")
        if self.scope:
            out.append(f"{pref}docid.scope:: {self.scope}")
        out.append(f"{pref}docid.id:: {self.id}")

        return "\n".join(out)

    def _remove_urn_part(self):
        self.id = re.sub(
            r"^(urn:iso:std:[^:]+"  # ISO prefix and originator
            r"  (?::(?:data|guide|isp|iwa|pas|r|tr|ts|tta))"  # type
            r"  ?:\d+)"  # docnumber
            r"(?::-[^:]+)?"  # partnumber
            r"(?::(draft|cancelled|stage-[^:]+))?"  # status
            r"(?::ed-\d+)?(?::v[^:]+)?"  # edition and version
            r"(?::\w{2}(?:,\w{2})*)?",  # langauge
            r"\1",
            self.id)
        self.id = re.sub(
            r"^(urn:iec:std:[^:]+"  # IEC prefix and originator
            r"  :\d+)"  # docnumber
            r"(?:-[^:]+)?",  # partnumber
            r"\1",
            self.id)



import dataclasses
import inspect
import pytest

import xml.etree.ElementTree as ET

from relaton_bib import Place


@pytest.fixture
def subject():
    return Place(name="name", region="FR")


def test_to_xml(subject):
    host = ET.Element("host")
    result = subject.to_xml(host)

    assert result.attrib["region"] == "FR"
    assert host.find("./place").text == "name"


def test_to_asciibib(subject):
    result = subject.to_asciibib()

    assert result == inspect.cleandoc("""place.name:: name
                                         place.region:: FR""")


def test_to_asciibib_with_pref(subject):
    result = subject.to_asciibib("m")

    assert result == inspect.cleandoc("""m.place.name:: name
                                         m.place.region:: FR""")


def test_hash(subject):
    result = dataclasses.asdict(subject)

    assert result["name"] == "name"
    assert result["region"] == "FR"



from dataclasses import dataclass

import xml.etree.ElementTree as ET

from .formatted_string import FormattedString


@dataclass(frozen=True)
class FormattedRef(FormattedString):
    def to_xml(self, parent):
        name = "formattedref"
        node = ET.Element(name) if parent is None \
            else ET.SubElement(parent, name)
        return super().to_xml(node)

    def to_asciibib(self, prefix=""):
        pref = f"{prefix}.formattedref" if prefix else "formattedref"
        return super().to_asciibib(pref)



from __future__ import annotations
from dataclasses import dataclass, field
from typing import Union, List

import xml.sax.saxutils as saxutils
import xml.etree.ElementTree as ET

from .relaton_bib import to_ds_instance


@dataclass(frozen=True)
class LocalizedString:
    content: Union[str, list[LocalizedString]]
    language: List[str] = field(default_factory=list)
    script: List[str] = field(default_factory=list)

    def __post_init__(self):
        inv = []
        if isinstance(self.content, list):
            def reject(x):
                return not isinstance(x, (LocalizedString, dict, str))
            inv = list(filter(reject, self.content))

        if not (isinstance(self.content, str) or not inv
                and (self.content and any(self.content))):
            klass = type(inv[0]) if isinstance(self.content, list) \
                else type(self.content)
            klass = klass.__name__
            raise ValueError(f"invalid LocalizedString content type: {klass}")

        # modify froozen dataclass https://stackoverflow.com/a/54119384/902217
        if isinstance(self.language, str):
            object.__setattr__(self, "language", [self.language])

        if isinstance(self.script, str):
            object.__setattr__(self, "script", [self.script])

        if isinstance(self.content, list):
            object.__setattr__(self, "content",
                               list(map(to_ds_instance(LocalizedString),
                                        self.content)))

    def __format__(self, format_spec: str) -> str:
        return str(self)

    def __str__(self):
        return self.content if isinstance(self.content, str) \
                            else str(self.content[0] if self.content else None)

    def __len__(self):
        return len(self.content)

    def empty(self):
        return len(self) == 0

    def to_xml(self, node):
        if not self.content:
            return

        if isinstance(self.content, list):
            for c in self.content:
                c.to_xml(ET.SubElement(node, "variant"))
        else:
            if any(self.language):
                node.attrib["language"] = ",".join(filter(None, self.language))
            if any(self.script):
                node.attrib["script"] = ",".join(filter(None, self.script))
            node.text = saxutils.escape(self.content)

        return node

    # TODO revisit
    # def to_hash
    #   if content.is_a? String
    #     return content unless language || script

    #     hash = { "content" => content }
    #     hash["language"] = single_element_array(language) if language&.any?
    #     hash["script"] = single_element_array(script) if script&.any?
    #     hash
    #   else content.map &:to_hash
    #   end
    # end

    def to_asciibib(self, prefix="", count=1, has_attrs=False):
        pref = f"{prefix}." if prefix else prefix

        if isinstance(self.content, list):
            return "\n".join(map(
                lambda c: c.to_asciibib(f"{pref}variant", len(self.content)),
                self.content))
        else:
            if not (any(self.language) or any(self.script) or has_attrs):
                return f"{prefix}:: {self.content}"

            out = [f"{prefix}::"] if count > 1 else []
            out.append(f"{pref}content:: {self.content}")
            for lang in self.language:
                out.append(f"{pref}language:: {lang}")
            for script in self.script:
                out.append(f"{pref}script:: {script}")
            return "\n".join(out)



from dataclasses import dataclass

import xml.etree.ElementTree as ET


@dataclass
class Place:
    name: str
    uri: str = None
    region: str = None

    def to_xml(self, parent):
        name = "place"
        node = ET.Element(name) if parent is None \
            else ET.SubElement(parent, name)
        node.text = self.name
        if self.uri:
            node.attrib["uri"] = self.uri
        if self.region:
            node.attrib["region"] = self.region
        return node

    def to_asciibib(self, prefix="", count=1):
        pref = f"{prefix}.place" if prefix else "place"
        out = [f"{pref}::"] if count > 1 else []
        out.append(f"{pref}.name:: {self.name}")
        if self.uri:
            out.append(f"{pref}.uri:: {self.uri}")
        if self.region:
            out.append(f"{pref}.region:: {self.region}")
        return "\n".join(out)



from __future__ import annotations
import dataclasses
import logging
import inspect
import pytest

import xml.etree.ElementTree as ET

from relaton_bib import FormattedString, ContributorRole, \
    ContributionInfo, Organization


@pytest.fixture
def role():
    return ContributorRole(type="type", description=[
        FormattedString("first"),
        FormattedString("second"),
        FormattedString("third")
    ])


@pytest.fixture
def info(role):
    return ContributionInfo(
        role=[role],
        entity=Organization(name="test"))


def test_role_invalid_type(caplog):
    invalid = "invalid"

    with caplog.at_level(logging.WARNING):
        ContributorRole(type=invalid)

    assert f"[relaton-bib] Contributor's type {invalid} is invalid" \
        in caplog.text


def test_role_to_xml(role):
    host = ET.Element("host")
    result = role.to_xml(host)

    assert result.attrib["type"] == "type"
    assert len(host.findall("./role/description")) == 3


def test_role_to_asciibib(role):
    result = role.to_asciibib()

    assert result == inspect.cleandoc(
        """role.description::
           role.description.content:: first
           role.description.format:: text/plain
           role.description::
           role.description.content:: second
           role.description.format:: text/plain
           role.description::
           role.description.content:: third
           role.description.format:: text/plain
           role.type:: type""")


def test_role_to_asciibib_with_pref(role):
    result = role.to_asciibib(prefix="test")

    assert result == inspect.cleandoc(
        """test.role.description::
           test.role.description.content:: first
           test.role.description.format:: text/plain
           test.role.description::
           test.role.description.content:: second
           test.role.description.format:: text/plain
           test.role.description::
           test.role.description.content:: third
           test.role.description.format:: text/plain
           test.role.type:: type""")


def test_role_hash(role):
    result = dataclasses.asdict(role)

    assert result["type"] == "type"
    assert len(result["description"]) == 3


def test_info_to_xml(info):
    host = ET.Element("host")
    result = info.to_xml(host)

    assert host.find("./organization/name").text == "test"


def test_info_to_asciibib(info):
    result = info.to_asciibib()

    assert result == inspect.cleandoc(
        """name:: test
           role.description::
           role.description.content:: first
           role.description.format:: text/plain
           role.description::
           role.description.content:: second
           role.description.format:: text/plain
           role.description::
           role.description.content:: third
           role.description.format:: text/plain
           role.type:: type""")


def test_info_to_asciibib_with_pref(info):
    result = info.to_asciibib(prefix="test")

    assert result == inspect.cleandoc(
        """test.name:: test
           test.role.description::
           test.role.description.content:: first
           test.role.description.format:: text/plain
           test.role.description::
           test.role.description.content:: second
           test.role.description.format:: text/plain
           test.role.description::
           test.role.description.content:: third
           test.role.description.format:: text/plain
           test.role.type:: type""")


def test_info_hash(info):
    result = dataclasses.asdict(info)

    assert result["entity"]["name"][0]["content"] == "test"
    assert len(result["role"][0]["description"]) == 3



import logging
import xml.etree.ElementTree as ET
import dateutil.parser as DU
from typing import List, Union

from .address import Address
from .affiliation import Affiliation
from .biblio_version import BibliographicItemVersion
from .bibliographic_date import BibliographicDate, BibliographicDateType
from .bibliographic_item import BibliographicItem
from .biblio_note import BiblioNoteCollection, BiblioNote
from .bib_item_locality import Locality, BibItemLocality, LocalityStack, \
    SourceLocalityStack, SourceLocality
from .contact import Contact, ContactType
from .classification import Classification
from .contribution_info import ContributionInfo, ContributorRole
from .copyright_association import CopyrightAssociation
from .document_identifier import DocumentIdentifier
from .document_relation import DocumentRelation
from .document_status import DocumentStatus
from .editorial_group import EditorialGroup
from .formatted_ref import FormattedRef
from .formatted_string import FormattedString, FormattedStringFormat
from .ics import ICS
from .localized_string import LocalizedString
from .person import Person, FullName, PersonIdentifier
from .place import Place
from .series import Series
from .structured_identifier import StructuredIdentifier, \
    StructuredIdentifierCollection
from .technical_committee import TechnicalCommittee
from .typed_title_string import TypedTitleString, TypedTitleStringCollection
from .typed_uri import TypedUri
from .organization import Organization, OrgIdentifier
from .medium import Medium
from .validity import Validity
from .workgroup import WorkGroup


def from_xml(xml: Union[ET.ElementTree, ET.Element]) -> BibliographicItem:
    bibitem = xml.getroot() if isinstance(xml, ET.ElementTree) else xml
    if bibitem.tag in ["bibitem", "bibdata"]:
        return _fetch_bibliographic_item(bibitem)
    else:
        logging.warning(
            "[relaton-bib] WARNING: "
            "can't find bibitem or bibdata element in the XML")


def _fetch_bibliographic_item(bibitem: ET.Element):
    fetched = bibitem.find("./fetched")
    if fetched is not None:
        fetched = DU.parse(fetched.text)
    docnumber = bibitem.find("./docnumber")
    if docnumber is not None:
        docnumber = docnumber.text
    edition = bibitem.find("./edition")
    if edition is not None:
        edition = edition.text
    ext = bibitem.find("./ext")
    doctype = None
    subdoctype = None
    if ext is not None:
        doctype = ext.find("doctype")
        if doctype is not None:
            doctype = doctype.text

        subdoctype = ext.find("subdoctype")
        if subdoctype is not None:
            subdoctype = subdoctype.text

    return BibliographicItem(
        id=bibitem.get("id", None),
        type=bibitem.get("type", None),
        fetched=fetched,
        title=_fetch_titles(bibitem),
        formattedref=_fref(bibitem),
        link=_fetch_link(bibitem),
        docidentifier=_fetch_docid(bibitem),
        docnumber=docnumber,
        date=_fetch_dates(bibitem),
        contributor=_fetch_contributors(bibitem),
        edition=edition,
        version=_fetch_version(bibitem),
        biblionote=_fetch_note(bibitem),
        language=_fetch_list(bibitem, "./language"),
        script=_fetch_list(bibitem, "./script"),
        abstract=_fetch_abstract(bibitem),
        status=_fetch_status(bibitem),
        copyright=_fetch_copyright(bibitem),
        relation=_fetch_relations(bibitem),
        series=_fetch_series(bibitem),
        medium=_fetch_medium(bibitem),
        place=_fetch_place(bibitem),
        extent=_fetch_extent(bibitem),
        accesslocation=_fetch_list(bibitem, "./accesslocation"),
        classification=_fetch_classification(bibitem),
        keyword=_fetch_list(bibitem, "./keyword"),
        license=_fetch_list(bibitem, "./license"),
        validity=_fetch_validity(bibitem),
        doctype=doctype,
        subdoctype=subdoctype,
        editorialgroup=_fetch_editorialgroup(ext),
        ics=_fetch_ics(ext),
        structuredidentifier=_fetch_structuredidentifier(ext),
    )


def _fetch_titles(item: ET.Element) -> TypedTitleStringCollection:
    return TypedTitleStringCollection(list(map(
        lambda t: _ttitle(t), item.findall("./title"))))


def _fetch_version(item: ET.Element) -> BibliographicItemVersion:
    version = item.find("./version")
    if not version:
        return

    rev_date = version.find("revision-date")
    if rev_date is not None:
        rev_date = rev_date.text
    draft = _fetch_list(version, "draft")

    return BibliographicItemVersion(revision_date=rev_date, draft=draft)


def _fetch_place(item: ET.Element) -> List[Place]:
    return list(map(
        lambda pl: Place(name=pl.text,
                         uri=pl.get("uri"),
                         region=pl.get("region")),
        item.findall("./place")))


def _fetch_note(item: ET.Element) -> BiblioNoteCollection:
    return BiblioNoteCollection(list(map(
        lambda n: BiblioNote(content=n.text, **n.attrib),
        item.findall("./note"))))


def _fetch_list(item: ET.Element, xpath: str) -> List[str]:
    if item is None:
        return []
    return list(map(
        lambda l: l.text,
        item.findall(xpath)))


def _fetch_series(item: ET.Element) -> List[Series]:
    result = []
    for sr in item.findall("./series"):
        abbr = sr.find("abbreviation")
        if abbr is not None:
            abbr = _localized_str(abbr)

        formattedref = _fref(sr)
        title = _ttitle(sr.find("title"))
        if not (formattedref or title):
            continue

        props = {p: n.text
                 for p in ["place",
                           "organization",
                           "from",
                           "to",
                           "number",
                           "partnumber"] if (n := sr.find(p)) is not None}

        result.append(Series(
            type=sr.get("type"),
            formattedref=formattedref,
            title=title,
            place=props.get("place"),
            organization=props.get("organization"),
            abbreviation=abbr,
            from_=props.get("from"),
            to=props.get("to"),
            number=props.get("number"),
            partnumber=props.get("partnumber")))
    return result


def _fetch_medium(item: ET.Element) -> Medium:
    medium = item.find("./medium")
    if not medium:
        return

    props = {p: n.text
             for p in ["form", "size", "scale"]
             if (n := medium.find(p)) is not None}

    return Medium(**props)


def _fetch_extent(item: ET.Element) -> List[BibItemLocality]:
    result = []

    for ext in item.findall("./extent"):
        type = ext.get("type")
        reference_from = ext.find("referenceFrom")
        if reference_from is not None:
            reference_from = reference_from.text
        reference_to = ext.find("referenceTo")
        if reference_to is not None:
            reference_to = reference_to.text
        result.append(BibItemLocality(
            type=type,
            reference_from=reference_from,
            reference_to=reference_to))

    return result


def _fetch_classification(item: ET.Element) -> List[Classification]:
    return list(map(
        lambda cls: Classification(type=cls.get("type"),
                                   value=cls.text),
        item.findall("./classification")))


def _fetch_validity(item: ET.Element) -> Validity:
    validity = item.find("./validity")
    if validity is None:
        return

    begins = validity.find("validityBegins")
    if begins is not None:
        begins = DU.parse(begins.text)
    ends = validity.find("validityEnds")
    if ends is not None:
        ends = DU.parse(ends.text)
    revision = validity.find("revision")
    if revision is not None:
        revision = DU.parse(revision.text)

    props = {p: DU.parse(n.text)
             for t, p in {"validityBegins": "begins",
                          "validityEnds": "ends",
                          "revision": "revision"}.items()
             if (n := validity.find(t)) is not None}

    return Validity(**props)


def _fetch_docid(item: ET.Element) -> List[DocumentIdentifier]:
    return list(map(
        lambda did: DocumentIdentifier(id=did.text,
                                       type=did.get("type"),
                                       scope=did.get("scope")),
        item.findall("./docidentifier")))


def _ttitle(title: ET.Element) -> TypedTitleString:
    if title is None:
        return []

    content = _localized_strs(title, "./variant")
    if not any(content):
        content = title.text

    props = title.attrib.copy()
    props["content"] = content

    return TypedTitleString(**props)


def _localized_strs(node: ET.Element, xpath: str) -> List[LocalizedString]:
    return list(map(
        lambda v: _localized_str(v),
        node.findall(xpath)))


def _fetch_status(item: ET.Element) -> DocumentStatus:
    status = item.find("./status")
    if status is None:
        return

    stg = status.find("stage")
    iter = status.find("iteration")
    if iter is not None:
        iter = iter.text
    return DocumentStatus(
        stage=status.text if stg is None else _stage(stg),
        substage=_stage(status.find("substage")),
        iteration=iter,
    )


def _stage(node: ET.Element) -> DocumentStatus.Stage:
    if node is None:
        return

    return DocumentStatus.Stage(
        value=node.text,
        abbreviation=node.get("abbreviation"))


def _fetch_dates(item: ET.Element) -> List[BibliographicDate]:
    result = []
    for d in item.findall("./date"):
        props = {p: n.text
                 for p in ["on", "from", "to"]
                 if (n := d.find(p)) is not None}
        props["type"] = d.get("type", BibliographicDateType.PUBLISHED)
        if "from" in props:
            props["from_"] = props.pop("from")
        elif "on" not in props:
            continue

        result.append(BibliographicDate(**props))

    return result


def _get_org(org: ET.Element) -> Organization:
    props = {p: n.text
             for p in ["abbreviation", "uri"]
             if (n := org.find(p)) is not None}
    props["name"] = list(map(
        lambda n: _localized_str(n),
        org.findall(f"./name")))
    props["identifier"] = list(map(
        lambda i: OrgIdentifier(value=i.text,
                                type=i.get("type")),
        org.findall(f"./identifier")))

    props["subdivision"] = _fetch_list(org, "subdivision")
    return Organization(**props)


def _get_person(person: ET.Element) -> Person:
    affiliations = []
    for a in person.findall("./affiliation"):
        desc = list(map(
            lambda d: _formatted_str(d), a.findall("./description")))
        affiliations.append(
            Affiliation(
                organization=_get_org(a.find("./organization")),
                description=desc
            )
        )

    contact = []
    for c in list(person):
        if c.tag == ContactType.ADDRESS:
            props = {p: n.text
                     for p in ["city", "state", "country", "postcode"]
                     if (n := c.find(p)) is not None}
            props["street"] = _fetch_list(c, "./street")
            contact.append(Address(**props))
        elif c.tag in [ContactType.PHONE,
                       ContactType.EMAIL,
                       ContactType.URI]:
            contact.append(Contact(type=c.tag, value=c.text))

    identifier = list(map(
        lambda pi: PersonIdentifier(type=pi.get("type"), value=pi.text),
        person.findall("./identifier")))

    fullname_props = dict(
        initial=_name_part(person, "initial"),
        forename=_name_part(person, "forename"),
        addition=_name_part(person, "addition"),
        prefix=_name_part(person, "prefix"))

    if (cname := person.find("./name/completename")) is not None:
        fullname_props["completename"] = _localized_str(cname)
    if (sname := person.find("./name/surname")) is not None:
        fullname_props["surname"] = _localized_str(sname)

    name = FullName(**fullname_props)

    return Person(
        name=name,
        affiliation=affiliations,
        contact=contact,
        identifier=identifier)


def _name_part(person: ET.Element, part: str) -> List[LocalizedString]:
    return list(map(
        lambda v: _localized_str(v),
        person.findall(f"./name/{part}")))


def _fetch_contributors(item: ET.Element) -> List[ContributionInfo]:
    result = []
    for c in item.findall("./contributor"):
        entity = None
        if (org := c.find("./organization")) is not None:
            entity = _get_org(org)
        elif (person := c.find("./person")) is not None:
            entity = _get_person(person)

        role = list(map(
            lambda r: ContributorRole(
                type=r.get("type"),
                description=_localized_strs(r, "./description")),
            c.findall("./role")))
        result.append(ContributionInfo(entity=entity, role=role))
    return result


def _fetch_abstract(item: ET.Element) -> List[FormattedString]:
    return list(map(lambda a: _formatted_str(a), item.findall("./abstract")))


def _fetch_copyright(item: ET.Element) -> List[TypedUri]:
    result = []
    for cp in item.findall("./copyright"):
        props = {p: n.text
                 for p in ["from", "to", "scope"]
                 if (n := cp.find(p)) is not None}
        props["from_"] = props.pop("from")
        props["owner"] = list(map(
            lambda o: ContributionInfo(
                entity=_get_org(o.find("organization"))),
            cp.findall("owner")))

        result.append(CopyrightAssociation(**props))
    return result


def _fetch_link(item: ET.Element) -> List[TypedUri]:
    return list(map(
        lambda l: TypedUri(type=l.get("type"), content=l.text),
        item.findall("./uri")))


def _fetch_relations(item: ET.Element, klass=DocumentRelation):
    result = []
    for rel in item.findall("./relation"):
        result.append(klass(
            type=rel.get("type"),
            description=_relation_description(rel),
            bibitem=_fetch_bibliographic_item(rel.find("./bibitem")),
            locality=_localities(rel),
            source_locality=_source_localities(rel),
        ))
    return result


def _relation_description(rel: ET.Element) -> FormattedString:
    d = rel.find("./description")
    if d is None:
        return

    return _formatted_str(d)


def _formatted_str(node: ET.Element) -> FormattedString:
    return FormattedString(content=node.text,
                           language=node.get("language", []),
                           script=node.get("script", []),
                           format=node.get("format",
                                           FormattedStringFormat.TEXT_PLAIN))


def _localized_str(node: ET.Element) -> LocalizedString:
    return LocalizedString(content=node.text,
                           language=node.get("language", []),
                           script=node.get("script", []))


def _localities(loc: ET.Element) -> List[LocalityStack]:
    result = []
    for lc in list(loc):
        if lc.tag not in ["locality", "localityStack"]:
            continue
        lcs = None
        if lc.get("type"):
            lcs = [_locality(lc)]
        else:
            lcs = list(filter(None, map(
                lambda l: _locality(l), lc.findall("./locality"))))
        result.append(LocalityStack(lcs))
    return result


def _locality(loc: ET.Element, klass=Locality):
    to = None
    if (rt := loc.find("./referenceTo")) is not None:
        to = LocalizedString(rt.text)
    fr0m = None
    if (rf := loc.find("./referenceFrom")) is not None:
        fr0m = LocalizedString(rf.text)

    return klass(
        type=loc.get("type"),
        reference_from=fr0m,
        reference_to=to
    )


def _source_localities(rel: ET.Element) -> List[SourceLocalityStack]:
    result = []
    for lc in list(rel):
        if lc.tag not in ["sourceLocality", "sourceLocalityStack"]:
            continue
        sls = None
        if lc.get("type"):
            sls = [_locality(lc, SourceLocality)]
        else:
            sls = list(filter(None, map(
                lambda l: _locality(l, SourceLocality),
                lc.findall("./sourceLocality"))))
        result.append(SourceLocalityStack(sls))
    return result


def _fref(item: ET.Element) -> FormattedRef:
    if not item:
        return
    ident = item.find("./formattedref")
    if ident is None:
        return

    return FormattedRef(
        content=ident.text,
        format=ident.get("format", FormattedStringFormat.TEXT_PLAIN),
        language=ident.get("language", []),
        script=ident.get("script", []))


def _fetch_editorialgroup(ext: ET.Element) -> EditorialGroup:
    if ext is None:
        return
    eg = ext.find("editorialgroup")
    if eg is None:
        return

    return EditorialGroup(list(map(
        lambda tc: TechnicalCommittee(
            WorkGroup(name=tc.text,
                      number=int(tc.get("number")),
                      type=tc.get("type"),
                      identifier=tc.get("identifier"),
                      prefix=tc.get("prefix"))),
        eg.findall("./technical-committee"))))


def _fetch_ics(ext: ET.Element) -> List[ICS]:
    if ext is None:
        return []

    result = []
    for ics in ext.findall("ics"):
        props = {p: n.text
                 for p in ["code", "text"]
                 if (n := ics.find(p)) is not None}
        result.append(ICS(**props))
    return result


def _fetch_structuredidentifier(ext: ET.Element) -> StructuredIdentifier:
    if ext is None:
        return

    sids = []
    for si in ext.findall("structuredidentifier"):
        agency = _fetch_list(si, "agency")
        class_ = si.find("class")
        if class_ is not None:
            class_ = class_.text
        props = {p: n.text
                 for p in ["docnumber",
                           "partnumber",
                           "edition",
                           "version",
                           "supplementtype",
                           "supplementnumber",
                           "language",
                           "year"] if (n := si.find(p)) is not None}
        sids.append(StructuredIdentifier(
            type=si.get("type"),
            agency=agency,
            class_=class_,
            docnumber=props.get("docnumber"),
            partnumber=props.get("partnumber"),
            edition=props.get("edition"),
            version=props.get("version"),
            supplementtype=props.get("supplementtype"),
            supplementnumber=props.get("supplementnumber"),
            language=props.get("language"),
            year=props.get("year")))
    return StructuredIdentifierCollection(sids)



import re
import datetime
from typing import Union, Dict, List

import dateutil.parser as DU

from .address import Address
from .affiliation import Affiliation
from .biblio_version import BibliographicItemVersion
from .bibliographic_date import BibliographicDate
from .bibliographic_item import BibliographicItem
from .biblio_note import BiblioNoteCollection, BiblioNote
from .bib_item_locality import Locality, BibItemLocality, LocalityStack, \
    SourceLocalityStack, SourceLocality
from .contact import Contact
from .classification import Classification
from .contribution_info import ContributionInfo, ContributorRole
from .copyright_association import CopyrightAssociation
from .document_identifier import DocumentIdentifier
from .document_relation import DocumentRelation
from .document_status import DocumentStatus
from .editorial_group import EditorialGroup
from .formatted_ref import FormattedRef
from .formatted_string import FormattedString
from .ics import ICS
from .localized_string import LocalizedString
from .person import Person, FullName, PersonIdentifier
from .place import Place
from .series import Series
from .structured_identifier import StructuredIdentifier, \
    StructuredIdentifierCollection
from .technical_committee import TechnicalCommittee
from .typed_title_string import TypedTitleString, TypedTitleStringCollection
from .typed_uri import TypedUri
from .organization import Organization, OrgIdentifier
from .medium import Medium
from .validity import Validity
from .workgroup import WorkGroup

from .relaton_bib import dict_replace_key


def from_dict(item: Dict) -> BibliographicItem:
    if not isinstance(item, dict):
        return None

    return BibliographicItem(
        fetched=item.get("fetched", datetime.datetime.now()),
        id=item.get("id"),
        type=item.get("type"),
        docidentifier=_docid(item),
        docnumber=item.get("docnumber"),
        script=_array(item.get("script")),
        language=_array(item.get("language")),
        version=_version(item),
        date=_dates(item),
        title=_titles(item),
        contributor=_contributors(item),
        abstract=_abstract(item),
        copyright=_copyright(item),
        link=[TypedUri(**link) for link in _array(item.get("link"))],
        ics=[ICS(**ics) for ics in _array(item.get("ics"))],
        keyword=_array(item.get("keyword")),
        accesslocation=_array(item.get("accesslocation")),
        place=_places(item),
        extent=_extent(item),
        biblionote=_notes(item),
        formattedref=_formattedref(item.get("formattedref")),
        status=_docstatus(item),
        relation=_relations(item),
        series=_series(item),
        medium=_medium(item),
        classification=_classification(item),
        validity=_validity(item),
        license=_array(item.get("license")),
        editorialgroup=_editorialgroup(item),
        structuredidentifier=_structuredidentifiers(item)
    )


def _array(arr: Union[List, Dict, None]) -> List:
    if not arr:
        return []
    elif not isinstance(arr, List):
        return [arr]

    return arr


def _extent(dictitem: Dict) -> List[BibItemLocality]:
    extent = _array(dictitem.get("extent"))
    return [BibItemLocality(reference_from=e.get("reference_from"),
                            reference_to=e.get("reference_to"),
                            type=e.get("type")) for e in extent]


def _titles(dictitem: Dict) -> TypedTitleStringCollection:
    result = []
    for title in _array(dictitem.get("title")):
        if isinstance(title, Dict):
            result.append(TypedTitleString(**title))
        elif isinstance(title, str):
            result.append(TypedTitleString.from_string(title))

    return TypedTitleStringCollection(result) if result else None


def _dates(dictitem: Dict) -> List[BibliographicDate]:
    return [
        BibliographicDate(**dict_replace_key(
            date,
            {"value": "on", "from": "from_"}
            )) for date in _array(dictitem.get("date"))
        ]


def _version(dictitem: Dict) -> BibliographicItemVersion:
    revdate = dictitem.get("revdate")
    return BibliographicItemVersion(revision_date=revdate) if revdate else None


def _abstract(dictitem: Dict) -> List[FormattedString]:
    # return [FormattedString(**a) for a in _array(dictitem.get("abstract"))]
    result = []
    for title in _array(dictitem.get("abstract")):
        if isinstance(title, str):
            result.append(FormattedString(content=title))
        else:
            result.append(FormattedString(**title))

    return result


def _places(dictitem: Dict) -> List[Place]:
    result = []
    for pl in _array(dictitem.get("place")):
        if isinstance(pl, str):
            result.append(Place(name=pl))
        else:
            result.append(Place(**pl))

    return result


def _docid(dictitem: Dict) -> List[Place]:
    result = []
    for docid in _array(dictitem.get("docid")):
        type_search = re.search(r"^\w+(?=\s)", docid.get("id"))
        type = docid.get("type", type_search.group(0) if type_search else None)
        result.append(
            DocumentIdentifier(id=docid.get("id"),
                               type=type,
                               scope=docid.get("scope"))
        )
    return result


def _notes(dictitem: Dict) -> List[BiblioNote]:
    result = []
    for note in _array(dictitem.get("biblionote")):
        if isinstance(note, str):
            result.append(BiblioNote(content=note))
        else:
            result.append(BiblioNote(**note))

    return BiblioNoteCollection(result) if result else None


def _docstatus(dictitem: Dict) -> DocumentStatus:
    if "docstatus" in dictitem:
        docstatus = dictitem.get("docstatus")
        return DocumentStatus(
            stage=_stage(docstatus.get("stage")),
            substage=_stage(docstatus.get("substage")),
            iteration=docstatus.get("iteration")
        )
    return None


def _stage(stg: Union[str, Dict]) -> DocumentStatus.Stage:
    if not stg:
        return None

    args = {"value": stg} if isinstance(stg, str) else stg
    return DocumentStatus.Stage(**args)


def _contributors(dictitem: Dict) -> List[ContributionInfo]:
    contrib = dictitem.get("contributor")
    result = []
    for c in _array(contrib):
        roles = []
        for r in _array(c.get("role")):
            if isinstance(r, Dict):
                roles.append(ContributorRole(
                    type=r.get("type"),
                    description=_array(r.get("description"))))
            else:
                roles.append(ContributorRole(type=r))

        entity = None
        if "person" in c:
            entity = _person(c.get("person"))
        else:
            entity = _org(c.get("organization"))

        result.append(ContributionInfo(entity=entity, role=roles))

    return result


def _org(org: Dict) -> Organization:
    if not org:
        return None

    org["identifier"] = [OrgIdentifier(
                         **dict_replace_key(idft, {"id": "value"}))
                         for idft in _array(org.get("identifier"))]
    org["subdivision"] = [_localizedstring(subd)
                          for subd in _array(org.get("subdivision"))]

    return Organization(**dict_replace_key(org, {"url": "uri"}))


def _person(person: Dict) -> Person:
    return Person(
        name=_fullname(person),
        affiliation=_affiliation(person),
        contact=_contacts(person),
        identifier=_person_identifiers(person)
    )


def _fullname(person) -> FullName:
    n = person.get("name")
    return FullName(
        forename=[_localname(f, person) for f in _array(n.get("forename"))],
        initial=[_localname(f, person) for f in _array(n.get("initial"))],
        addition=[_localname(f, person) for f in _array(n.get("addition"))],
        prefix=[_localname(f, person) for f in _array(n.get("prefix"))],
        surname=_localname(n.get("surname"), person),
        completename=_localname(n.get("completename"), person))


def _person_identifiers(person: Dict) -> List[PersonIdentifier]:
    return [PersonIdentifier(**dict_replace_key(a, {"id": "value"}))
            for a in _array(person.get("identifier"))]


def _affiliation(person: Dict) -> List[Affiliation]:
    affiliation = person.get("affiliation")
    if not affiliation:
        return []

    result = []
    for a in _array(affiliation):
        descr = []
        for d in _array(a.get("description")):
            cnt = d
            if not isinstance(d, Dict):
                cnt = {"content": d}
            descr.append(FormattedString(**cnt))
        result.append(Affiliation(
            organization=_org(a.get("organization")),
            description=descr
        ))

    return result


def _contacts(person: Dict) -> List[Contact]:
    contact = person.get("contact")
    if not contact:
        return []

    result = []
    for a in _array(contact):
        if a.get("city") or a.get("country"):
            a["street"] = _array(a.pop("street"))
            result.append(Address(**a))
        else:
            result.append(Contact(**a))

    return result


def _copyright(dictitem: Dict) -> List[CopyrightAssociation]:
    result = []

    for c in _array(dictitem.get("copyright")):
        if "owner" in c:
            owner = []
            for o in _array(c.get("owner")):
                entity = None
                if "person" in c:
                    entity = _person(c.get("person"))
                else:
                    entity = _org(c.get("organization"))
                owner.append(ContributionInfo(entity=entity))
            c["owner"] = owner

        result.append(CopyrightAssociation(
            from_=DU.parse(c.get("from")),
            owner=c.get("owner")
        ))

    return result


def _relations(dictitem: Dict) -> List[DocumentRelation]:
    relations = dictitem.get("relation")
    if not relations:
        return []

    result = []
    for r in _array(relations):
        if r.get("description"):
            r["description"] = FormattedString(**r.get("description"))

        result.append(DocumentRelation(
            type=r.get("type"),
            bibitem=_relation_bibitem(r),
            description=r.get("description"),
            locality=_relation_locality(r),
            source_locality=_relation_source_locality(r)))

    return result


def _relation_bibitem(dictitem: Dict) -> DocumentRelation:
    return from_dict(dictitem["bibitem"]) if "bibitem" in dictitem else None


def _relation_locality(rel: Dict) -> List[LocalityStack]:
    result = []
    for bl in _array(rel.get("locality")):
        locality = None
        if "locality_stack" in bl:
            locality = [Locality(**ls)
                        for ls in _array(bl.get("locality_stack"))]
        else:
            locality = [Locality(**bl)]
        result.append(LocalityStack(locality=locality))
    return result


def _relation_source_locality(rel: Dict) -> List[SourceLocalityStack]:
    result = []
    src_locality = rel.get("source_locality")
    for s in _array(src_locality):
        src_locs = None
        if "source_locality_stack" in s:
            src_locs = [SourceLocality(**loc)
                        for loc in src_locality["source_locality_stack"]]
        else:
            src_locs = [SourceLocality(**s)]
        result.append(SourceLocalityStack(locality=src_locs))
    return result


def _series(dictitem: Dict) -> List[Series]:
    result = []
    for s in _array(dictitem.get("series")):
        if "formattedref" in s:
            s["formattedref"] = _formattedref(s.pop("formattedref"))

        if "title" in s:
            if not isinstance(s["title"], Dict):
                s["title"] = {"content": s.pop("title")}
            s["title"] = _typed_string(s.pop("title"))

        if "abbreviation" in s:
            s["abbreviation"] = _localizedstring(s.pop("abbreviation"))

        result.append(Series(**dict_replace_key(s, {"from": "from_"})))

    return result


def _typed_string(title: Dict) -> TypedTitleString:
    return TypedTitleString(**title)


def _medium(dictitem: Dict) -> Medium:
    return Medium(**dictitem.get("medium")) if dictitem.get("medium") else None


def _classification(dictitem: Dict) -> List[Classification]:
    return [Classification(**c)
            for c in _array(dictitem.get("classification"))]


def _validity(dictitem: Dict) -> Validity:
    if dictitem is None or dictitem.get("validity") is None:
        return None

    validity = dictitem.get("validity")

    args = dict(
        begins=_parse_validity_time(validity, "begins"),
        ends=_parse_validity_time(validity, "ends"),
        revision=_parse_validity_time(validity, "revision")
    )

    return Validity(**args)


def _parse_validity_time(val: Dict, period: str) -> datetime.datetime:
    if isinstance(val.get(period), datetime.datetime):
        return val.get(period)

    return DU.parse(val.get(period))


def _editorialgroup(dictitem: Dict) -> EditorialGroup:
    result = [TechnicalCommittee(WorkGroup(**wg))
              for wg in _array(dictitem.get("editorialgroup"))]

    return EditorialGroup(result) if result else None


def _ics(dictitem: Dict) -> List[ICS]:
    return [ICS(**ics) for ics in dictitem.get("ics")]


def _structuredidentifiers(dictitem: Dict) -> StructuredIdentifierCollection:
    result = []
    for si in _array(dictitem.get("structuredidentifier")):
        si["agency"] = _array(si.pop("agency"))
        result.append(StructuredIdentifier(
            **dict_replace_key(si, {"class": "class_"})))

    return StructuredIdentifierCollection(result) if result else None


def _localname(name: Dict, person: Dict) -> LocalizedString:
    if not name:
        return None

    result = name if isinstance(name, Dict) else {}
    if isinstance(person.get("name"), Dict):
        pname = person.get("name")
        if "language" not in result and pname.get("language"):
            result["language"] = pname.get("language")
        if "script" not in result and pname.get("script"):
            result["script"] = pname.get("script")
        result["content"] = name.get("content") if isinstance(name, Dict) \
            else name

    return LocalizedString(**result)


def _localizedstring(lst: Union[str, List, Dict]) -> LocalizedString:
    return LocalizedString(**lst) if isinstance(lst, Dict) \
        else LocalizedString(content=lst)


def _formattedref(frf: Union[str, Dict]) -> FormattedRef:
    if not frf:
        return None

    return FormattedRef(**frf) if isinstance(frf, Dict) \
        else FormattedRef(content=frf)



from dataclasses import dataclass
from enum import Enum

import logging
import xml.etree.ElementTree as ET


class ContactType(str, Enum):
    PHONE = "phone"
    EMAIL = "email"
    URI = "uri"
    ADDRESS = "address"

    @classmethod
    def has_value(cls, value):
        return value in cls._value2member_map_


@dataclass
class Contact:
    type: str
    value: str

    def __post_init__(self):
        if not ContactType.has_value(self.type):
            logging.warning(
                f"[relaton-bib] invalid contact type: {self.type}")

    def to_xml(self, parent):
        result = ET.Element(self.type) if parent is None \
            else ET.SubElement(parent, self.type)
        result.text = self.value
        return result

    def to_asciibib(self, prefix="", count=1):
        pref = f"{prefix}." if prefix else prefix
        out = [f"{pref}contact::"] if count > 1 else []
        out.append(f"{pref}contact.type:: {self.type}")
        out.append(f"{pref}contact.value:: {self.value}")
        return "\n".join(out)


import pytest

import xml.etree.ElementTree as ET

from relaton_bib import LocalizedString


def test_missin_props():
    with pytest.raises(ValueError) as excinfo:
        LocalizedString(None)
    assert "invalid LocalizedString content type: NoneType" \
        in str(excinfo.value)


def test_not_empty():
    loc_str = LocalizedString(
        "content & character to escape", "en", "Latn")
    assert len(loc_str) != 0


def test_xml_ampersand_escaped():
    loc_str = LocalizedString(
        "content & character to escape", "en", "Latn")

    host = ET.Element("host")
    result = loc_str.to_xml(host)

    assert result.attrib["language"] == "en"
    assert result.attrib["script"] == "Latn"

    xmlstr = ET.tostring(result, encoding='unicode', method='xml')

    assert "content &amp;amp; character to escape" in xmlstr



from __future__ import annotations
from dataclasses import dataclass, field
from multiprocessing import Pool
from typing import List

import xml.etree.ElementTree as ET

from .relaton_bib import delegate


def hit_fetch(hit):
    return hit.fetch()


@dataclass
@delegate("array", "append", "__getitem__", "__len__", "__iter__",
          "__reversed__", "__contains__")
class HitCollection:
    text: str
    year: str = None
    fetched: bool = False
    array: List = field(default_factory=list)

    def fetch(self):
        workers = Pool(4)
        result = workers.map(hit_fetch, self.array)
        # TODO looks incomplete
        self.fetched = True
        return self

    # @param opts [Hash]
    # @option opts [Nokogiri::XML::Builder] :builder XML builder
    # @option opts [Boolean] :bibdata
    # @option opts [String, Symbol] :lang language
    # @return [String] XML
    def to_xml(self, parent=None, opts={}):
        name = "documents"
        result = ET.Element(name) if parent is None \
            else ET.SubElement(parent, name)

        for hit in self.array:
            hit.fetch()
            hit.to_xml(result, opts)

        return result
